"""Tasks."""

from datetime import timedelta
from typing import Optional

from celery import shared_task

from django.contrib.auth.models import User
from django.db import DatabaseError, transaction
from django.utils.timezone import now
from esi.decorators import rate_limit_retry_task

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from structuretimers import __title__
from structuretimers.models import (
    DiscordWebhook,
    DistancesFromStaging,
    NotificationRule,
    ScheduledNotification,
    StagingSystem,
    Timer,
)

logger = get_extension_logger(__name__)
TASK_PRIORITY_HIGH = 4


@shared_task(base=QueueOnce, acks_late=True)
def send_messages_for_webhook(webhook_pk: int) -> None:
    """Send all currently queued messages for given webhook to Discord."""
    webhook = DiscordWebhook.objects.get(pk=webhook_pk)
    if not webhook.is_enabled:
        logger.info("Tracker %s: DiscordWebhook disabled - skipping sending", webhook)
        return
    logger.info("Started sending messages to webhook %s", webhook)
    webhook.send_queued_messages()
    logger.info("Completed sending messages to webhook %s", webhook)


@shared_task(base=QueueOnce, bind=True, acks_late=True)
def send_scheduled_notification(self, scheduled_notification_pk: int) -> None:
    """Send a scheduled notification for a timer based on a notification rule."""
    with transaction.atomic():
        try:
            scheduled_notification = (
                ScheduledNotification.objects.select_for_update().get(
                    pk=scheduled_notification_pk
                )
            )
        except (ScheduledNotification.DoesNotExist, DatabaseError):
            logger.info(
                "ScheduledNotification with pk = %s does not / no longer exist "
                "or is being processed by another task. Discarding.",
                scheduled_notification_pk,
            )
            logger.debug(
                "ScheduledNotification pk = %s. Debug info",
                scheduled_notification_pk,
                exc_info=True,
            )
            return

        logger.debug(
            "Delete scheduled_notification in task_id = %s: %r",
            self.request.id,
            scheduled_notification,
        )
        scheduled_notification.delete()

    if scheduled_notification.celery_task_id != self.request.id:
        logger.info(
            "Discarded outdated scheduled notification: %r", scheduled_notification
        )
        return

    notification_rule = scheduled_notification.notification_rule
    if not notification_rule.is_enabled:
        logger.info(
            "Discarded scheduled notification based on disabled rule: %r",
            scheduled_notification,
        )
        return

    webhook = notification_rule.webhook
    if not webhook.is_enabled:
        logger.warning(
            "Webhook not enabled for %r. Discarding.", scheduled_notification
        )
        return

    timer = scheduled_notification.timer
    if (
        not timer.date
        or timer.date < now()
        or scheduled_notification.timer_date < now()
    ):  # fix issue #28
        logger.warning(
            "Discarding scheduled notification %r for outdated timer.",
            scheduled_notification,
        )
        return

    logger.info(
        "Sending notifications for timer '%s' and rule '%s'",
        timer,
        notification_rule,
    )
    minutes = round((timer.date - now()).total_seconds() / 60)
    mod_text = "**important** " if timer.is_important else ""
    content = (
        f"The following {mod_text}structure timer will elapse "
        f"in less than **{minutes:,}** minutes:"
    )
    timer.send_notification(
        webhook=webhook,
        content=notification_rule.prepend_ping_text(content),
    )
    send_messages_for_webhook.apply_async(
        args=[webhook.pk], priority=TASK_PRIORITY_HIGH
    )


@shared_task
def notify_about_new_timer(timer_pk: int, notification_rule_pk: int) -> None:
    """Send notification about new timer."""
    timer = Timer.objects.get(pk=timer_pk)
    notification_rule = NotificationRule.objects.select_related("webhook").get(
        pk=notification_rule_pk
    )
    if not notification_rule.is_enabled or not notification_rule.webhook.is_enabled:
        return

    author_text = f" by **{timer.eve_character}**" if timer.eve_character else ""
    content = f"New timer added{author_text}:"
    timer.send_notification(
        webhook=notification_rule.webhook,
        content=notification_rule.prepend_ping_text(content),
    )
    send_messages_for_webhook.apply_async(
        args=[notification_rule.webhook.pk], priority=TASK_PRIORITY_HIGH
    )


@shared_task(acks_late=True)
def schedule_notifications_for_timer(timer_pk: int, is_new: bool = False) -> None:
    """Schedule notifications for this timer based on notification rules."""
    timer = Timer.objects.select_related_for_matching().get(pk=timer_pk)
    if not timer.date:
        raise ValueError(f"Not supported for preliminary timers: {timer}")

    if timer.date < now():
        logger.warning("Can not schedule notification for past timer: %s", timer)
        return

    # trigger: newly created
    if is_new:
        rules = (
            NotificationRule.objects.select_related("webhook")
            .filter(
                is_enabled=True,
                trigger=NotificationRule.Trigger.NEW_TIMER_CREATED,
                webhook__is_enabled=True,
            )
            .conforms_with_timer(timer)
        )
        if rules:
            for rule in rules:
                notify_about_new_timer.apply_async(
                    kwargs={"timer_pk": timer.pk, "notification_rule_pk": rule.pk},
                    priority=TASK_PRIORITY_HIGH,
                )

    # trigger: timer elapses soon
    with transaction.atomic():
        # remove existing scheduled notifications if date has changed
        for obj in timer.scheduled_notifications.exclude(timer_date=timer.date):
            _revoke_notification_for_timer(scheduled_notification=obj)

        # schedule new notifications
        for notification_rule in NotificationRule.objects.filter(
            is_enabled=True,
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
        ).conforms_with_timer(timer):
            _schedule_notification_for_timer(
                timer=timer, notification_rule=notification_rule
            )


@shared_task(acks_late=True)
def schedule_notifications_for_rule(notification_rule_pk: int) -> None:
    """Schedule notifications for all timers confirming with this rule.

    Will recreate all existing and still pending notifications
    """
    notification_rule = NotificationRule.objects.get(pk=notification_rule_pk)
    if notification_rule.trigger == NotificationRule.Trigger.NEW_TIMER_CREATED:
        logger.error(
            "NotificationRule with pk = %s has the wrong trigger. Aborting.",
            notification_rule_pk,
        )
        return

    logger.debug("Checking scheduled notifications for: %s", notification_rule)
    with transaction.atomic():
        for obj in notification_rule.scheduled_notifications.filter(
            timer_date__gt=now()
        ):
            _revoke_notification_for_timer(scheduled_notification=obj)

        for timer in Timer.objects.filter(
            date__gt=now()
        ).conforms_with_notification_rule(notification_rule):
            _schedule_notification_for_timer(
                timer=timer, notification_rule=notification_rule
            )


def _schedule_notification_for_timer(
    timer: Timer, notification_rule: NotificationRule
) -> ScheduledNotification:
    """Schedule notification for a timer."""
    if timer.timer_type == Timer.Type.PRELIMINARY:
        raise ValueError(f"Can not schedule preliminary timers: {timer}")
    if not timer.date:
        raise ValueError(f"Timer has no date: {timer}")
    if not notification_rule.scheduled_time:
        raise ValueError(
            f"Notification rule has no scheduled date: {notification_rule}"
        )
    logger.info(
        "Scheduling fresh notification for timer #%d, rule #%d",
        timer.pk,
        notification_rule.pk,
    )
    notification_date = timer.date - timedelta(minutes=notification_rule.scheduled_time)
    scheduled_notification, _ = ScheduledNotification.objects.update_or_create(
        timer=timer,
        notification_rule=notification_rule,
        defaults={"timer_date": timer.date, "notification_date": notification_date},
    )
    result = send_scheduled_notification.apply_async(
        kwargs={"scheduled_notification_pk": scheduled_notification.pk},
        eta=timer.date - timedelta(minutes=notification_rule.scheduled_time),
        priority=TASK_PRIORITY_HIGH,
    )
    scheduled_notification.celery_task_id = result.task_id
    scheduled_notification.save()

    return scheduled_notification


def _revoke_notification_for_timer(
    scheduled_notification: ScheduledNotification,
) -> None:
    """Revoke notification for a timer."""
    logger.info(
        "Removing stale notification for timer #%d, rule #%d",
        scheduled_notification.timer.pk,
        scheduled_notification.notification_rule.pk,
    )
    scheduled_notification.delete()


@shared_task
def send_test_message_to_webhook(
    webhook_pk: int, user_pk: Optional[int] = None
) -> None:
    """Send a test message to given webhook.
    Optionally inform user about result if user ok is given
    """
    webhook = DiscordWebhook.objects.get(pk=webhook_pk)
    user = User.objects.get(pk=user_pk) if user_pk else None
    logger.info("Sending test message to webhook %s", webhook)
    error_text, success = webhook.send_test_message(user)

    if not user:
        return

    message = (
        f"Error text: {error_text}\nCheck log files for details."
        if not success
        else "No errors"
    )
    level = "success" if success else "error"
    notify(
        user=user,
        title=(
            f"{__title__}: Result of test message to webhook {webhook}: "
            f"{level.upper()}"
        ),
        message=message,
        level=level,
    )


@shared_task
def housekeeping() -> None:
    """Perform housekeeping tasks"""
    logger.info("Performing housekeeping")
    deleted_count = Timer.objects.delete_obsolete()
    logger.info("Deleted %d obsolete timers.", deleted_count)


@shared_task
def calc_staging_system(staging_system_pk: int, force_update: bool = False) -> None:
    """Recalc distances from a staging system for all timers."""
    for timer_pk in Timer.objects.values_list("pk", flat=True):
        calc_timer_distances_for_staging_system.delay(
            timer_pk, staging_system_pk, force_update
        )


@shared_task
def calc_timer_distances_for_all_staging_systems(
    timer_pk: int, force_update: bool = False
) -> None:
    """Recalc distances for a timer from a staging system."""
    timer = Timer.objects.get(pk=timer_pk)
    for staging_system_pk in StagingSystem.objects.values_list("pk", flat=True):
        calc_timer_distances_for_staging_system.apply_async(
            kwargs={
                "timer_pk": timer.pk,
                "staging_system_pk": staging_system_pk,
                "force_update": force_update,
            },
            priority=TASK_PRIORITY_HIGH,
        )


@shared_task
@rate_limit_retry_task
def calc_timer_distances_for_staging_system(
    timer_pk: int, staging_system_pk: int, force_update: bool = False
) -> None:
    """Calc distances for a timer from a staging system."""
    timer = Timer.objects.get(pk=timer_pk)
    staging_system = StagingSystem.objects.get(pk=staging_system_pk)
    DistancesFromStaging.objects.calc_timer_for_staging_system(
        timer=timer, staging_system=staging_system, force_update=force_update
    )



from io import StringIO
from unittest import skipIf
from unittest.mock import Mock, patch

from django.core.management import call_command
from eveuniverse.tests.testdata.factories_2 import EveTypeFactory

from allianceauth.timerboard.models import Timer as AuthTimer
from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from structuretimers.constants import EveTypeId
from structuretimers.management.commands.structuretimers_migrate_timers import (
    _STRUCTURE_MAP,
)
from structuretimers.models import Timer
from structuretimers.tests.testdata.factory import (
    AuthTimerFactory,
    EveSolarSystemLowSecFactory,
    UserWithAccessFactory,
)
from structuretimers.tests.utils import isolated_subtest

PACKAGE_PATH = "structuretimers.management.commands"
MODELS_PATH = "structuretimers.models"


@skipIf("timerboard" not in app_labels(), "timerboard not installed")
@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
@patch(MODELS_PATH + ".STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
class TestMigrateTimers(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.system_amamake = EveSolarSystemLowSecFactory(id=30002537, name="Amamake")
        cls.user = UserWithAccessFactory()

        for name, type_id in _STRUCTURE_MAP.items():
            EveTypeFactory(id=type_id, name=name)

        Timer.objects.all().delete()
        AuthTimer.objects.all().delete()

    def setUp(self) -> None:
        self.out = StringIO()

    def test_can_convert_all_fields(self):
        # given
        character_1 = self.user.profile.main_character
        corporation_1 = self.user.profile.main_character.corporation
        auth_timer = AuthTimerFactory(
            eve_character=character_1,
            eve_corp=corporation_1,
            objective=AuthTimer.Objective.FRIENDLY,
            structure=AuthTimer.Structure.ASTRAHUS,
            timer_type=AuthTimer.TimerType.ARMOR,
            system="Amamake",
            user=self.user,
            planet_moon="moon location",
        )

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)

        # then
        self.assertEqual(Timer.objects.count(), 1)
        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.date, auth_timer.eve_time)
        self.assertEqual(new_timer.details_notes, auth_timer.details)
        self.assertEqual(new_timer.eve_character, character_1)
        self.assertEqual(new_timer.eve_corporation, corporation_1)
        self.assertEqual(new_timer.eve_solar_system, self.system_amamake)
        self.assertEqual(new_timer.location_details, auth_timer.planet_moon)
        self.assertEqual(new_timer.objective, Timer.Objective.FRIENDLY)
        self.assertEqual(new_timer.structure_type.id, EveTypeId.ASTRAHUS)
        self.assertEqual(new_timer.timer_type, Timer.Type.ARMOR)
        self.assertEqual(new_timer.user, self.user)

    def test_should_handle_all_timer_types(self):
        for timer_type in AuthTimer.TimerType.values:
            with self.subTest(timer_type=timer_type), isolated_subtest():
                # given
                AuthTimerFactory(
                    timer_type=timer_type,
                    user=self.user,
                )

                # when
                call_command("structuretimers_migrate_timers", stdout=self.out)

                # then
                self.assertEqual(Timer.objects.count(), 1, msg=timer_type)

    def test_should_handle_all_objectives(self):
        for objective in AuthTimer.Objective.values:
            with self.subTest(objective=objective), isolated_subtest():
                # given
                AuthTimerFactory(
                    objective=objective,
                    user=self.user,
                )

                # when
                call_command("structuretimers_migrate_timers", stdout=self.out)

                # then
                self.assertEqual(Timer.objects.count(), 1, msg=objective)

    def test_should_handle_all_structure_types(self):
        supported_values = set(AuthTimer.Structure.values) - {AuthTimer.Structure.OTHER}
        for structure in supported_values:
            with self.subTest(structure=structure), isolated_subtest():
                # given
                AuthTimerFactory(
                    structure=structure,
                    user=self.user,
                )

                # when
                call_command("structuretimers_migrate_timers", stdout=self.out)

                # then
                self.assertEqual(Timer.objects.count(), 1, msg=structure)

    def test_final_corp_timer(self):
        # given
        AuthTimerFactory(
            corp_timer=True,
            user=self.user,
        )

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)

        # then
        self.assertEqual(Timer.objects.count(), 1)
        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.visibility, Timer.Visibility.CORPORATION)

    def test_moon_mining(self):
        # given
        AuthTimerFactory(
            structure=AuthTimer.Structure.MOONPOP,
            timer_type=AuthTimer.TimerType.UNSPECIFIED,
            user=self.user,
        )

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)

        # then
        new_timer = Timer.objects.first()
        self.assertEqual(new_timer.timer_type, Timer.Type.MOONMINING)
        self.assertEqual(new_timer.structure_type.id, EveTypeId.ATHANOR)

    def test_should_skip_timers_with_unknown_system(self):
        # given
        AuthTimerFactory(
            system="Invalid",
            user=self.user,
        )

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)

        self.assertFalse(Timer.objects.all().exists())

    def test_should_skip_timers_with_unknown_structure(self):
        # given
        AuthTimerFactory(
            structure="Invalid",
            user=self.user,
        )

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)

        self.assertFalse(Timer.objects.all().exists())

    def test_should_not_create_duplicates(self):
        # given
        AuthTimerFactory()

        # when
        call_command("structuretimers_migrate_timers", stdout=self.out)
        call_command("structuretimers_migrate_timers", stdout=self.out)

        # then
        self.assertEqual(Timer.objects.all().count(), 1)



# Generated by Django 4.0.3 on 2022-04-19 18:58

import multiselectfield.db.fields

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("eveonline", "0015_factions"),
        ("structuretimers", "0003_add_preliminary_timers"),
    ]

    operations = [
        migrations.AddField(
            model_name="notificationrule",
            name="exclude_regions",
            field=models.ManyToManyField(
                blank=True,
                help_text="Timer must NOT be created within one of the given regions",
                related_name="+",
                to="eveuniverse.everegion",
            ),
        ),
        migrations.AddField(
            model_name="notificationrule",
            name="exclude_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space Type must NOT be one of the selected",
                max_length=11,
            ),
        ),
        migrations.AddField(
            model_name="notificationrule",
            name="require_regions",
            field=models.ManyToManyField(
                blank=True,
                help_text="Timer must be created within one of the given regions or leave blank to match any.",
                related_name="+",
                to="eveuniverse.everegion",
            ),
        ),
        migrations.AddField(
            model_name="notificationrule",
            name="require_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space type must be one of the selected or leave blank to match any.",
                max_length=11,
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="exclude_alliances",
            field=models.ManyToManyField(
                blank=True,
                help_text="Timer must NOT be created by one of the given alliances",
                related_name="+",
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_alliances",
            field=models.ManyToManyField(
                blank=True,
                help_text="Timer must be created by one of the given alliances or leave blank to match any.",
                related_name="+",
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_corporations",
            field=models.ManyToManyField(
                blank=True,
                help_text="Timer must be created by one of the given corporations or leave blank to match any.",
                related_name="+",
                to="eveonline.evecorporationinfo",
            ),
        ),
    ]



# Generated by Django 4.2.16 on 2024-11-04 14:50

import multiselectfield.db.fields

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structuretimers", "0005_alter_notificationrule_exclude_space_types_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notificationrule",
            name="exclude_timer_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                ],
                help_text="Timer must NOT have one of the given timer types",
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_timer_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                ],
                help_text="Timer must have one of the given timer types or leave blank to match any.",
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="timer",
            name="timer_type",
            field=models.CharField(
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                    ("PL", "Preliminary"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
    ]



from django.test import TestCase
from eveuniverse.tests.testdata.factories_2 import EveTypeFactory

from structuretimers.constants import EveCategoryId, EveGroupId, EveTypeId
from structuretimers.selectors import supported_eve_types


class TestSupportedEveTypes(TestCase):
    def test_should_return_supported_types(self):
        # given
        structure = EveTypeFactory(eve_group__eve_category__id=EveCategoryId.STRUCTURE)
        control_tower = EveTypeFactory(
            eve_group__id=EveGroupId.CONTROL_TOWER, published=True
        )
        mobile_depot = EveTypeFactory(
            eve_group__id=EveGroupId.MOBILE_DEPOT, published=True
        )
        mercenary_den = EveTypeFactory(
            eve_group__id=EveGroupId.MERCENARY_DEN, published=True
        )
        pirate_fob = EveTypeFactory(
            eve_group__id=EveGroupId.PIRATE_FORWARD_OPERATING_BASE
        )
        customs_office = EveTypeFactory(id=EveTypeId.CUSTOMS_OFFICE)
        orbital_skyhook = EveTypeFactory(id=EveTypeId.ORBITAL_SKYHOOK)
        ihub = EveTypeFactory(id=EveTypeId.IHUB)
        tcu = EveTypeFactory(id=EveTypeId.TCU)

        EveTypeFactory(eve_group__eve_category__id=25)  # Asteroid - should be ignored

        # when
        got = supported_eve_types()

        # then
        self.assertCountEqual(
            got,
            [
                structure,
                control_tower,
                mobile_depot,
                mercenary_den,
                pirate_fob,
                customs_office,
                orbital_skyhook,
                ihub,
                tcu,
            ],
        )



import datetime as dt
import json
from typing import NamedTuple
from unittest.mock import Mock, patch

import dhooks_lite

from django.core.cache import cache
from django.test import TestCase, override_settings
from django.utils.timezone import now
from eveuniverse.models import EveSolarSystem

from app_utils.json import JSONDateTimeDecoder
from app_utils.testing import NoSocketsTestCase

from structuretimers import __title__
from structuretimers.models import (
    ScheduledNotification,
    StagingSystem,
    Timer,
    _task_calc_staging_system,
)

from .testdata.factory import (
    CitadelTypeFactory,
    DiscordWebhookFactory,
    DistancesFromStagingFactory,
    EveSolarSystemFactory,
    EveSolarSystemHighSecFactory,
    EveSolarSystemLowSecFactory,
    EveSolarSystemNullSecFactory,
    EveSolarSystemWSpaceFactory,
    NotificationRuleFactory,
    ScheduledNotificationFactory,
    StagingSystemFactory,
    TimerFactory,
    UserWithAccessFactory,
)

MODULE_PATH = "structuretimers.models"


class TestTimer(NoSocketsTestCase):
    def test_str_1(self):
        timer = TimerFactory.build(
            structure_name="Test",
            timer_type=Timer.Type.ARMOR,
            eve_solar_system=EveSolarSystemFactory(name="Abune"),
            structure_type=CitadelTypeFactory(name="Astrahus"),
            date=dt.datetime(2020, 8, 6, 13, 25, tzinfo=dt.timezone.utc),
        )
        expected = 'Armor timer for Astrahus "Test" in Abune @ 2020-08-06 13:25'
        self.assertEqual(str(timer), expected)

    def test_str_2(self):
        timer = TimerFactory.build(
            structure_name="Test",
            timer_type=Timer.Type.PRELIMINARY,
            eve_solar_system=EveSolarSystemFactory(name="Abune"),
            structure_type=CitadelTypeFactory(name="Astrahus"),
        )
        expected = 'Preliminary timer for Astrahus "Test" in Abune'
        self.assertEqual(str(timer), expected)

    def test_structure_display_name_1(self):
        timer = TimerFactory.build(
            eve_solar_system=EveSolarSystemFactory(name="Abune"),
            structure_type=CitadelTypeFactory(name="Astrahus"),
        )
        expected = "Astrahus in Abune"
        self.assertEqual(timer.structure_display_name, expected)

    def test_structure_display_name_2(self):
        timer = TimerFactory.build(
            eve_solar_system=EveSolarSystemFactory(name="Abune"),
            structure_type=CitadelTypeFactory(name="Astrahus"),
            location_details="P5-M3",
        )
        expected = "Astrahus in Abune near P5-M3"
        self.assertEqual(timer.structure_display_name, expected)

    def test_structure_display_name_3(self):
        timer = TimerFactory.build(
            structure_name="Big Boy",
            eve_solar_system=EveSolarSystemFactory(name="Abune"),
            structure_type=CitadelTypeFactory(name="Astrahus"),
        )
        expected = 'Astrahus "Big Boy" in Abune'
        self.assertEqual(timer.structure_display_name, expected)


@patch(MODULE_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
class TestTimer_ScheduleNotificationsOnSave(NoSocketsTestCase):
    @patch(MODULE_PATH + "._task_schedule_notifications_for_timer")
    def test_schedule_notifications_for_new_timers(self, mock_schedule_notifications):
        # when
        timer = Timer.objects.create(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=EveSolarSystemFactory(),
            structure_type=CitadelTypeFactory(),
            timer_type=Timer.Type.ARMOR,
        )

        # then
        self.assertTrue(mock_schedule_notifications.called)
        _, kwargs = mock_schedule_notifications.return_value.apply_async.call_args
        self.assertEqual(kwargs["kwargs"]["timer_pk"], timer.pk)

    def test_schedule_notifications_when_date_changed(self):
        # when
        timer = TimerFactory(date=now() + dt.timedelta(hours=4))

        with patch(
            MODULE_PATH + "._task_schedule_notifications_for_timer"
        ) as mock_schedule_notifications:
            timer.date = now() + dt.timedelta(hours=3)
            timer.save()
            self.assertTrue(mock_schedule_notifications.called)
            _, kwargs = mock_schedule_notifications.return_value.apply_async.call_args
            self.assertEqual(kwargs["kwargs"]["timer_pk"], timer.pk)

    def test_dont_schedule_notifications_else(self):
        timer = TimerFactory(
            date=now() + dt.timedelta(hours=4),
        )

        with patch(
            MODULE_PATH + "._task_schedule_notifications_for_timer"
        ) as mock_schedule_notifications:
            timer.date = now() + dt.timedelta(hours=3)
            timer.structure_name = "Some fancy name"
            self.assertFalse(mock_schedule_notifications.called)

    @patch(MODULE_PATH + "._task_schedule_notifications_for_timer")
    def test_dont_schedule_notifications_for_new_preliminary_timers(
        self, mock_schedule_notifications
    ):
        # when
        Timer.objects.create(
            timer_type=Timer.Type.PRELIMINARY,
            eve_solar_system=EveSolarSystemFactory(),
            structure_type=CitadelTypeFactory(),
        )

        # then
        self.assertFalse(mock_schedule_notifications.called)

    @patch(MODULE_PATH + "._task_schedule_notifications_for_timer")
    def test_remove_scheduled_notifications_when_timer_changed_to_preliminary(
        self, mock_schedule_notifications
    ):
        # given
        rule = NotificationRuleFactory(is_enabled=False)
        timer = TimerFactory(date=now() + dt.timedelta(hours=4))
        notification = ScheduledNotificationFactory(notification_rule=rule, timer=timer)
        mock_schedule_notifications.reset()

        # when
        timer.timer_type = Timer.Type.PRELIMINARY
        timer.save()

        # then
        self.assertFalse(mock_schedule_notifications.called)
        self.assertFalse(
            ScheduledNotification.objects.filter(pk=notification.pk).exists()
        )


@patch(MODULE_PATH + "._task_schedule_notifications_for_timer", Mock)
class TestTimer_CalcDistancesOnSave(NoSocketsTestCase):
    @patch(MODULE_PATH + "._task_calc_timer_distances_for_all_staging_systems")
    def test_should_calc_distances_when_created(self, mock_calc_distances):
        # when
        timer = Timer.objects.create(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=EveSolarSystemFactory(),
            structure_type=CitadelTypeFactory(),
        )

        # then
        self.assertTrue(mock_calc_distances.called)
        _, kwargs = mock_calc_distances.return_value.apply_async.call_args
        self.assertEqual(kwargs["args"][0], timer.pk)

    @patch(MODULE_PATH + "._task_calc_timer_distances_for_all_staging_systems")
    def test_should_recalc_distances_when_solar_system_has_changed(
        self, mock_calc_distances
    ):
        # given
        timer: Timer = TimerFactory(date=now() + dt.timedelta(hours=4))

        # when
        timer.eve_solar_system = EveSolarSystemFactory()
        timer.save()

        # then
        self.assertTrue(mock_calc_distances.called)

    @patch(MODULE_PATH + "._task_calc_timer_distances_for_all_staging_systems")
    def test_should_not_recalc_distances_when_other_fields_changed(
        self, mock_calc_distances
    ):
        # given
        timer = TimerFactory(
            date=now() + dt.timedelta(hours=4),
        )
        # when
        timer.structure_type = CitadelTypeFactory()
        timer.save()

        # then
        self.assertFalse(mock_calc_distances.called)


class TestTimer_UserCanEdit(NoSocketsTestCase):
    def test_creator_can_edit_own_timer(self):
        user = UserWithAccessFactory(
            permissions__=[
                "structuretimers.basic_access",
                "structuretimers.create_timer",
            ]
        )
        timer = TimerFactory(user=user)
        self.assertTrue(timer.user_can_edit(user))

    def test_manager_can_edit_other_timers(self):
        user = UserWithAccessFactory(
            permissions__=[
                "structuretimers.basic_access",
                "structuretimers.manage_timer",
            ]
        )
        timer = TimerFactory(user=UserWithAccessFactory())
        self.assertTrue(timer.user_can_edit(user))

    def test_non_manager_can_not_edit_other_timer(self):
        user = UserWithAccessFactory(
            permissions__=[
                "structuretimers.basic_access",
                "structuretimers.create_timer",
            ]
        )
        timer = TimerFactory(user=UserWithAccessFactory())
        self.assertFalse(timer.user_can_edit(user))

    """
    def test_user_with_basic_access_can_view_normal_timer(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            user=self.user_1,
        )
        self.assertTrue(timer.user_can_view(self.user_3))

    def test_user_can_not_view_corp_restricted_timer_from_other_corp(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            eve_corporation=self.corporation_1,
            visibility=Timer.Visibility.CORPORATION,
            user=self.user_1,
        )
        self.assertFalse(timer.user_can_view(self.user_3))

    def test_user_can_view_corp_restricted_timer_from_same_corp(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            eve_corporation=self.corporation_1,
            visibility=Timer.Visibility.CORPORATION,
            user=self.user_1,
        )
        self.assertTrue(timer.user_can_view(self.user_2))

    def test_user_can_not_view_alliance_restricted_timer_from_other_alliance(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            eve_alliance=self.alliance_1,
            visibility=Timer.Visibility.ALLIANCE,
            user=self.user_1,
        )
        self.assertFalse(timer.user_can_view(self.user_3))

    def test_opsec_user_can_view_opsec_timer(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            is_opsec=True,
            user=self.user_2,
        )
        self.assertTrue(timer.user_can_view(self.user_2))

    def test_non_opsec_user_can_not_view_opsec_timer(self):
        timer = Timer(
            date=now() + dt.timedelta(hours=4),
            eve_solar_system=self.system_abune,
            structure_type=self.type_astrahus,
            is_opsec=True,
            user=self.user_2,
        )
        self.assertFalse(timer.user_can_view(self.user_1))
    """


@patch(MODULE_PATH + ".DiscordWebhook.send_message", spec=True)
class TestTimer_SendNotification(NoSocketsTestCase):
    @patch(MODULE_PATH + ".STRUCTURETIMER_NOTIFICATION_SET_AVATAR", True)
    def test_should_send_minimal_notification(self, mock_send_message):
        # given
        timer = TimerFactory()
        webhook = DiscordWebhookFactory()

        # when
        timer.send_notification(webhook)

        # then
        self.assertEqual(mock_send_message.call_count, 1)
        _, kwargs = mock_send_message.call_args
        self.assertEqual(kwargs["username"], __title__)
        self.assertIsNotNone(kwargs["avatar_url"])

    @patch(MODULE_PATH + ".STRUCTURETIMER_NOTIFICATION_SET_AVATAR", False)
    def test_should_send_notification_without_avatar(self, mock_send_message):
        # given
        timer = TimerFactory()
        webhook = DiscordWebhookFactory()

        # when
        timer.send_notification(webhook)

        # then
        self.assertEqual(mock_send_message.call_count, 1)
        _, kwargs = mock_send_message.call_args
        self.assertIsNone(kwargs["username"])
        self.assertIsNone(kwargs["avatar_url"])

    def test_with_content(self, mock_send_message):
        # given
        timer = TimerFactory()
        webhook = DiscordWebhookFactory()

        # when
        timer.send_notification(webhook, "Extra Text")

        # then
        self.assertEqual(mock_send_message.call_count, 1)
        _, kwargs = mock_send_message.call_args
        self.assertIn("Extra Text", kwargs["content"])

    def test_timer_with_options_1(self, mock_send_message):
        # given
        timer = TimerFactory(objective=Timer.Objective.FRIENDLY)
        webhook = DiscordWebhookFactory()

        # when
        timer.send_notification(webhook)

        # then
        self.assertEqual(mock_send_message.call_count, 1)

    def test_timer_with_options_2(self, mock_send_message):
        # given
        timer = TimerFactory(objective=Timer.Objective.HOSTILE)
        webhook = DiscordWebhookFactory()

        # when
        timer.send_notification(webhook)

        # then
        self.assertEqual(mock_send_message.call_count, 1)


class TestTimerSpaceType_FromEveSolarSystem(NoSocketsTestCase):
    def test_all(self):
        class Case(NamedTuple):
            name: str
            solar_system: EveSolarSystem
            want: Timer.SpaceType

        cases = [
            Case("high sec", EveSolarSystemHighSecFactory(), Timer.SpaceType.HIGH_SEC),
            Case("low sec", EveSolarSystemLowSecFactory(), Timer.SpaceType.LOW_SEC),
            Case("null sec", EveSolarSystemNullSecFactory(), Timer.SpaceType.NULL_SEC),
            Case("w sec", EveSolarSystemWSpaceFactory(), Timer.SpaceType.WH_SPACE),
        ]
        for tc in cases:
            got = Timer.SpaceType.from_eve_solar_system(tc.solar_system)
            self.assertEqual(got, tc.want)


class TestDiscordWebhook(TestCase):
    def test_str(self):
        webhook = DiscordWebhookFactory(name="Dummy")
        self.assertEqual(str(webhook), "Dummy")

    def test_repr(self):
        webhook = DiscordWebhookFactory(name="Dummy")
        self.assertEqual(
            repr(webhook), f"DiscordWebhook(id={webhook.id}, name='Dummy')"
        )

    def test_queue_features(self):
        cache.clear()
        webhook = DiscordWebhookFactory(name="Dummy")
        self.assertEqual(webhook.queue_size(), 0)
        webhook.send_message(content="Dummy message")
        self.assertEqual(webhook.queue_size(), 1)
        webhook.clear_queue()
        self.assertEqual(webhook.queue_size(), 0)

    def test_send_message_normal(self):
        cache.clear()
        webhook = DiscordWebhookFactory(name="Dummy")
        embed = dhooks_lite.Embed(description="my_description")
        self.assertEqual(
            webhook.send_message(
                content="my_content",
                username="my_username",
                avatar_url="my_avatar_url",
                embeds=[embed],
            ),
            1,
        )
        message = json.loads(webhook._main_queue.dequeue(), cls=JSONDateTimeDecoder)
        expected = {
            "content": "my_content",
            "embeds": [{"description": "my_description", "type": "rich"}],
            "username": "my_username",
            "avatar_url": "my_avatar_url",
        }
        self.assertDictEqual(message, expected)

    def test_send_message_empty(self):
        cache.clear()
        webhook = DiscordWebhookFactory(name="Dummy")
        with self.assertRaises(ValueError):
            webhook.send_message()


@patch(MODULE_PATH + ".sleep", new=lambda x: x)
@patch(MODULE_PATH + ".DiscordWebhook.send_message_to_webhook", spec=True)
class TestDiscordWebhook_SendQueuedMessages(NoSocketsTestCase):
    def setUp(self):
        cache.clear()

    def test_one_message(self, mock_send_message_to_webhook):
        """
        when one message in queue
        then send it and returns 1
        """
        mock_send_message_to_webhook.return_value = True
        webhook = DiscordWebhookFactory()
        webhook.send_message("dummy")

        result = webhook.send_queued_messages()

        self.assertEqual(result, 1)
        self.assertTrue(mock_send_message_to_webhook.called)
        self.assertEqual(webhook.queue_size(), 0)

    def test_three_message(self, mock_send_message_to_webhook):
        """
        when three messages in queue
        then sends them and returns 3
        """
        mock_send_message_to_webhook.return_value = True
        webhook = DiscordWebhookFactory()
        webhook.send_message("dummy-1")
        webhook.send_message("dummy-2")
        webhook.send_message("dummy-3")

        result = webhook.send_queued_messages()

        self.assertEqual(result, 3)
        self.assertEqual(mock_send_message_to_webhook.call_count, 3)
        self.assertEqual(webhook.queue_size(), 0)

    def test_no_messages(self, mock_send_message_to_webhook):
        """
        when no message in queue
        then do nothing and return 0
        """
        mock_send_message_to_webhook.return_value = True
        webhook = DiscordWebhookFactory()
        result = webhook.send_queued_messages()

        self.assertEqual(result, 0)
        self.assertFalse(mock_send_message_to_webhook.called)
        self.assertEqual(webhook.queue_size(), 0)

    def test_failed_message(self, mock_send_message_to_webhook):
        """
        given one message in queue
        when sending fails
        then re-queues message and return 0
        """
        mock_send_message_to_webhook.return_value = False
        webhook = DiscordWebhookFactory()
        webhook.send_message("dummy")

        result = webhook.send_queued_messages()

        self.assertEqual(result, 0)
        self.assertTrue(mock_send_message_to_webhook.called)
        self.assertEqual(webhook.queue_size(), 1)


@patch(MODULE_PATH + ".dhooks_lite.Webhook.execute", spec=True)
@patch(MODULE_PATH + ".logger", spec=True)
class TestDiscordWebhook_SendMessageToWebhook(NoSocketsTestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_send_normal(self, mock_logger, mock_execute):
        """
        when sending of message successful
        return True
        """
        mock_execute.return_value = dhooks_lite.WebhookResponse(
            headers={}, status_code=200
        )
        message = {
            "content": "my_content",
            "embeds": [{"description": "my_description", "type": "rich"}],
            "username": "my_username",
            "avatar_url": "my_avatar_url",
        }
        webhook = DiscordWebhookFactory()

        result = webhook.send_message_to_webhook(message)

        self.assertTrue(result)
        self.assertTrue(mock_execute.called)
        _, kwargs = mock_execute.call_args
        self.assertDictEqual(
            kwargs,
            {
                "content": "my_content",
                "embeds": [
                    dhooks_lite.Embed.from_dict(
                        {"description": "my_description", "type": "rich"}
                    )
                ],
                "username": "my_username",
                "avatar_url": "my_avatar_url",
                "wait_for_response": True,
            },
        )
        self.assertFalse(mock_logger.warning.called)

    def test_send_failed(self, mock_logger, mock_execute):
        """
        when sending of message failed
        then log warning and return False
        """
        mock_execute.return_value = dhooks_lite.WebhookResponse(
            headers={}, status_code=440
        )
        message = {
            "content": "my_content",
            "embeds": [{"description": "my_description", "type": "rich"}],
            "username": "my_username",
            "avatar_url": "my_avatar_url",
        }
        webhook = DiscordWebhookFactory()

        result = webhook.send_message_to_webhook(message)

        self.assertFalse(result)
        self.assertTrue(mock_execute.called)
        self.assertTrue(mock_logger.warning.called)


@patch(MODULE_PATH + ".EveSolarSystem.distance_to", lambda *args, **kwargs: 4.257e16)
@patch(MODULE_PATH + ".EveSolarSystem.jumps_to", lambda *args, **kwargs: 3)
@patch(MODULE_PATH + "._task_calc_staging_system", wraps=_task_calc_staging_system)
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestStagingSystem(NoSocketsTestCase):
    def test_should_calc_distances(self, spy_task_calc_staging_system):
        # given
        timer = TimerFactory(
            structure_name="Test",
            timer_type=Timer.Type.ARMOR,
            date=dt.datetime(2020, 8, 6, 13, 25, tzinfo=dt.timezone.utc),
        )

        # when
        staging_system = StagingSystem.objects.create(
            eve_solar_system=EveSolarSystemFactory()
        )

        # then
        obj = timer.distances.first()
        self.assertEqual(obj.staging_system, staging_system)
        self.assertAlmostEqual(obj.light_years, 4.5, delta=0.1)
        self.assertEqual(obj.jumps, 3)
        self.assertTrue(spy_task_calc_staging_system.called)

    def test_should_not_update_distances_when_solar_system_not_changed(
        self, spy_task_calc_staging_system
    ):
        # given
        TimerFactory()
        staging_system = StagingSystemFactory()

        # when
        staging_system.save()

        # then
        self.assertFalse(spy_task_calc_staging_system.called)


@patch(MODULE_PATH + ".EveSolarSystem.jumps_to", spec=True)
@patch(MODULE_PATH + ".EveSolarSystem.distance_to", spec=True)
class TestDistancesFromStaging_Calculate(NoSocketsTestCase):
    def test_should_calculate_distances(self, mock_distance_to, mock_jumps_to):
        # given
        mock_distance_to.return_value = 2.3
        mock_jumps_to.return_value = 4
        timer = TimerFactory()
        staging_system = StagingSystemFactory()
        distances = DistancesFromStagingFactory(
            timer=timer,
            staging_system=staging_system,
            light_years=None,
            jumps=None,
        )
        # when
        distances.calculate()
        # then
        self.assertGreater(distances.light_years, 0)
        self.assertEqual(distances.jumps, 4)

    def test_should_calculate_distances_when_none(
        self, mock_distance_to, mock_jumps_to
    ):
        # given
        mock_distance_to.return_value = 2.3
        mock_jumps_to.return_value = 4
        timer = TimerFactory()
        staging_system = StagingSystemFactory(eve_solar_system=None)
        distances = DistancesFromStagingFactory(
            timer=timer,
            staging_system=staging_system,
            light_years=None,
            jumps=None,
        )
        # when
        distances.calculate()
        # then
        self.assertIsNone(distances.light_years)
        self.assertIsNone(distances.jumps)



"""Views."""

# pylint: disable=too-many-ancestors,missing-function-docstring, missing-class-docstring

import math
from copy import deepcopy
from typing import Iterable

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.views import View
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    ListView,
    TemplateView,
    UpdateView,
)
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.eveonline.evelinks import dotlan
from allianceauth.services.hooks import get_extension_logger
from app_utils.views import (
    JSONResponseMixin,
    fontawesome_link_button_html,
    link_html,
    yesno_str,
)

from structuretimers import __title__
from structuretimers.app_settings import (
    STRUCTURETIMERS_DEFAULT_PAGE_LENGTH,
    STRUCTURETIMERS_PAGING_ENABLED,
)
from structuretimers.forms import TimerForm
from structuretimers.models import DistancesFromStaging, StagingSystem, Timer
from structuretimers.selectors import supported_eve_types

logger = get_extension_logger(__name__)
DATETIME_FORMAT = "%Y-%m-%d %H:%M"
MAX_HOURS_PASSED = 2


def bootstrap5_label_html(text: str, label: str = "secondary") -> str:
    """Return HTML for a Bootstrap 5 label."""
    return format_html('<span class="badge text-bg-{}">{}</span>', label, text)


class TimerListView(LoginRequiredMixin, PermissionRequiredMixin, TemplateView):
    """View for showing a list of timer."""

    template_name = "structuretimers/timer_list.html"
    permission_required = "structuretimers.basic_access"

    def get_context_data(self, **kwargs):
        staging_systems_qs = StagingSystem.objects.select_related(
            "eve_solar_system", "eve_solar_system__eve_constellation__eve_region"
        ).filter(eve_solar_system__isnull=False)
        selected_staging_system = None
        staging_system_name = self.request.GET.get("staging")
        if staging_system_name:
            try:
                selected_staging_system = staging_systems_qs.get(
                    eve_solar_system__name=self.request.GET.get("staging")
                )
            except (StagingSystem.DoesNotExist, ValueError):
                pass
        if not selected_staging_system:
            selected_staging_system = staging_systems_qs.filter(is_main=True).first()
            if not selected_staging_system:
                selected_staging_system = staging_systems_qs.first()
        stageing_systems = staging_systems_qs.order_by("eve_solar_system__name")
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "current_time": now().strftime("%H:%M"),
                "max_hours_expired": MAX_HOURS_PASSED,
                "title": __title__,
                "data_tables_page_length": STRUCTURETIMERS_DEFAULT_PAGE_LENGTH,
                "data_tables_paging": STRUCTURETIMERS_PAGING_ENABLED,
                "selected_staging_system": selected_staging_system,
                "stageing_systems": stageing_systems,
                "tab": self.request.GET.get("tab", "current"),
            }
        )
        return context


class TimerListDataView(
    LoginRequiredMixin, PermissionRequiredMixin, JSONResponseMixin, ListView
):
    """Produce timer list in JSON for AJAX call."""

    model = Timer
    permission_required = "structuretimers.basic_access"

    def render_to_response(self, context, **response_kwargs):
        return self.render_to_json_response(context, **response_kwargs)

    def get_queryset(self):
        qs = super().get_queryset()
        timers_qs = qs.visible_to_user(self.request.user)
        timers_qs = timers_qs.filter_by_tab(
            tab_name=self.kwargs.get("tab_name"), max_hours_passed=MAX_HOURS_PASSED
        )
        timers_qs = timers_qs.select_related(
            "eve_solar_system",
            "eve_solar_system__eve_constellation__eve_region",
            "structure_type",
            "structure_type__eve_group",
            "eve_character",
            "eve_corporation",
            "eve_alliance",
        )
        return timers_qs

    def get_data(self, context):
        data = []
        timers: Iterable[Timer] = self.object_list
        for timer in timers:
            location = self._calc_location_for_timer(timer)
            distances, distance_text = self._calc_distance_for_timer(timer)
            structure = self._calc_structure_for_timer(timer)
            is_restricted, objective = self._calc_objective(timer)
            owner_name, name = self._calc_owner_name(timer)
            visibility = self._calc_visibility(timer)
            distances_light_years = distances.light_years if distances else None
            data.append(
                {
                    "id": timer.id,
                    "local_time": timer.date.isoformat() if timer.date else "",
                    "date": timer.date.isoformat() if timer.date else "",
                    "location": location,
                    "structure_details": structure,
                    "name_objective": name,
                    "owner": objective,
                    # "creator": creator,
                    "distance": {
                        "display": distance_text,
                        "sort": distances_light_years,
                    },
                    "distance_light_years": distances_light_years,
                    "distance_jumps": distances.jumps if distances else None,
                    "actions": self._get_data_actions(timer),
                    "timer_type_name": timer.get_timer_type_display(),
                    "objective_name": timer.get_objective_display(),
                    "system_name": timer.eve_solar_system.name,
                    "region_name": timer.eve_solar_system.eve_constellation.eve_region.name,
                    "structure_type_name": timer.structure_type.name,
                    "owner_name": owner_name,
                    "visibility": visibility,
                    "opsec_str": yesno_str(timer.is_opsec),
                    "is_opsec": timer.is_opsec,
                    "is_passed": timer.date < now() if timer.date else None,
                    "is_important": timer.is_important,
                    "is_restricted": is_restricted,
                    "last_updated_at": timer.last_updated_at.isoformat(),
                }
            )

        return data

    def _calc_visibility(self, timer):
        if timer.eve_corporation:
            corporation_name = timer.eve_corporation.corporation_name
        else:
            corporation_name = "-"

        visibility = ""
        if timer.visibility == Timer.Visibility.ALLIANCE and timer.eve_alliance:
            visibility = timer.eve_alliance.alliance_name
        elif timer.visibility == Timer.Visibility.CORPORATION:
            visibility = corporation_name
        return visibility

    def _calc_owner_name(self, timer):
        if timer.owner_name:
            owner_name = timer.owner_name
            owner = owner_name
        else:
            owner = "-"
            owner_name = ""

        structure_name = timer.structure_name if timer.structure_name else "-"
        name = format_html("{}<br>{}", structure_name, owner)
        return owner_name, name

    def _calc_objective(self, timer):
        tags = []
        is_restricted = False
        if timer.is_opsec:
            tags.append(bootstrap5_label_html("OPSEC", "danger"))
            is_restricted = True

        if timer.visibility != Timer.Visibility.UNRESTRICTED:
            tags.append(bootstrap5_label_html(timer.get_visibility_display(), "info"))
            is_restricted = True

        if timer.is_important:
            tags.append(bootstrap5_label_html("Important", "warning"))

        objective = format_html(
            "{}<br>{}",
            mark_safe(
                bootstrap5_label_html(
                    timer.get_objective_display(), timer.label_type_for_objective()
                )
            ),
            mark_safe(" ".join(tags)),
        )

        return is_restricted, objective

    def _calc_structure_for_timer(self, timer):
        if timer.structure_type:
            structure_type_icon_url = timer.structure_type.icon_url(size=64)
            structure_type_name = timer.structure_type.name
        else:
            structure_type_icon_url = ""
            structure_type_name = "(unknown)"

        context = {
            "type_icon_url": structure_type_icon_url,
            "type_name": structure_type_name,
            "timer_name": timer.get_timer_type_display(),
            "timer_style": timer.label_type_for_timer_type(),
        }
        return render_to_string("structuretimers/partials/structure_box.html", context)

    @staticmethod
    def _calc_location_for_timer(timer: Timer):
        location = link_html(
            dotlan.solar_system_url(timer.eve_solar_system.name),
            timer.eve_solar_system.name,
        )
        if timer.location_details:
            location += format_html("<br><em>{}</em>", timer.location_details)

        location += format_html(
            "<br>{}", timer.eve_solar_system.eve_constellation.eve_region.name
        )
        return location

    def _calc_distance_for_timer(self, timer: Timer):
        staging_system_pk = self.request.GET.get("staging")
        if staging_system_pk:
            distances_map = {
                obj.timer_id: obj
                for obj in DistancesFromStaging.objects.filter(
                    staging_system__pk=staging_system_pk
                ).all()
            }
        else:
            distances_map = {}
        try:
            distances = distances_map[timer.id]
        except KeyError:
            distance_text = "?"
            distances = None
        else:
            light_years_text = (
                f"{math.ceil(distances.light_years * 10) / 10} ly"
                if distances.light_years is not None
                else "N/A"
            )
            jumps_text = (
                f"{distances.jumps} jumps" if distances.jumps is not None else "N/A"
            )
            distance_text = format_html("{}<br>{}", light_years_text, jumps_text)
        return distances, distance_text

    def _get_data_actions(self, timer: Timer):
        actions = ""
        if timer.details_image_url or timer.details_notes:
            disabled_html = ""
            button_type = "primary"
            data_toggle = 'data-bs-toggle="modal" data-bs-target="#modalTimerDetails" '
            title = "Show details of this timer"
        else:
            button_type = "secondary"
            disabled_html = " disabled"
            data_toggle = ""
            title = "No details available"
        actions += (
            format_html(
                '<button type="button" id="timerboardBtnDetails" '
                'class="btn btn-{}" title="{}"'
                "{}"
                'data-timerpk="{}"{}>'
                '<i class="fas fa-search-plus"></i>'
                "</button>",
                button_type,
                title,
                mark_safe(data_toggle),
                timer.pk,
                mark_safe(disabled_html),
            )
            + "&nbsp;"
        )
        if timer.user_can_edit(self.request.user):
            actions += (
                fontawesome_link_button_html(
                    reverse("structuretimers:delete", args=(timer.pk,)),
                    "far fa-trash-alt",
                    "danger",
                    "Delete this timer",
                )
                + "&nbsp;"
                + fontawesome_link_button_html(
                    reverse("structuretimers:edit", args=(timer.pk,)),
                    "far fa-edit",
                    "warning",
                    "Edit this timer",
                )
            )
        if self.request.user.has_perm("structuretimers.create_timer"):
            actions += "&nbsp;" + fontawesome_link_button_html(
                reverse("structuretimers:copy", args=(timer.pk,)),
                "far fa-copy",
                "success",
                "Copy this timer",
            )
        return actions


class TimerDetailDataView(LoginRequiredMixin, PermissionRequiredMixin, DetailView):
    """View for showing details of a timer."""

    permission_required = "structuretimers.basic_access"
    model = Timer

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.visible_to_user(self.request.user).select_related(
            "structure_type", "eve_solar_system"
        )


class TimerManagementView(LoginRequiredMixin, PermissionRequiredMixin, View):
    """View for editing a timer."""

    model = Timer
    form_class = TimerForm
    title = _("Edit Structure Timer")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = self.title
        return context

    def send_success_message(self, keyword: str) -> None:
        """Inform user about result of his action via message."""
        timer = self.object
        messages.info(self.request, f"{keyword}: {timer}.")


class AddUpdateMixin:
    def get_form_kwargs(self):
        """Inject the request user into the kwargs passed to the form."""
        kwargs = super().get_form_kwargs()
        kwargs.update({"user": self.request.user})
        return kwargs


class CreateTimerView(TimerManagementView, AddUpdateMixin, CreateView):
    template_name_suffix = "_create_form"
    permission_required = (
        "structuretimers.basic_access",
        "structuretimers.create_timer",
    )
    title = _("Create New Timer")

    def form_valid(self, form):
        result = super().form_valid(form)
        timer = self.object
        logger.info(
            "Created new timer in %s at %s by user %s",
            timer.eve_solar_system,
            timer.date,
            self.request.user,
        )
        self.send_success_message(_("Added"))
        return result


# FIXME: User without permission can still post to edit timers directly


class EditTimerMixin:
    permission_required = "structuretimers.basic_access"

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if response.status_code == 200:
            can_view = (
                Timer.objects.filter(pk=self.object.pk)
                .visible_to_user(self.request.user)
                .exists()
            )
            can_edit = self.object.user_can_edit(self.request.user)
            if not can_view or not can_edit:
                raise PermissionDenied()

        return response


class EditTimerView(EditTimerMixin, TimerManagementView, AddUpdateMixin, UpdateView):
    template_name_suffix = "_update_form"

    def form_valid(self, form):
        result = super().form_valid(form)
        self.send_success_message(_("Updated"))
        return result


class CopyTimerView(CreateTimerView):
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        old_obj = get_object_or_404(Timer, pk=self.kwargs["pk"])
        new_obj = deepcopy(old_obj)
        new_obj.pk = None
        new_obj.date = None
        kwargs["instance"] = deepcopy(new_obj)
        return kwargs


# FIXME: User without permission can still post to delete timers directly


class RemoveTimerView(
    EditTimerMixin, LoginRequiredMixin, PermissionRequiredMixin, DeleteView
):
    model = Timer

    def get_success_url(self) -> str:
        return self.object.get_absolute_url()


class Select2SolarSystemsView(JSONResponseMixin, ListView):
    """Dynamically generated list of solar systems for select2 widget."""

    model = EveSolarSystem

    def get_queryset(self):
        qs = super().get_queryset()
        term = self.request.GET.get("term")
        if not term:
            return qs.none()
        return qs.filter(name__istartswith=term)

    def get_context_data(self, **kwargs):
        if self.object_list:
            solar_systems = self.object_list.values("id", "name")
            results = [{"id": row["id"], "text": row["name"]} for row in solar_systems]
            results = sorted(results, key=lambda d: d["text"])
        else:
            results = None
        return {"results": results}

    def render_to_response(self, context, **response_kwargs):
        return self.render_to_json_response(context, **response_kwargs)


class Select2StructureTypesView(JSONResponseMixin, ListView):
    """Dynamically generated list of types for select2 widget."""

    def get_queryset(self):
        term = self.request.GET.get("term")
        if not term:
            return EveType.objects.none()

        qs = supported_eve_types()
        return qs.filter(name__icontains=term)

    def get_context_data(self, **kwargs):
        if self.object_list:
            results = [
                {"id": row["id"], "text": row["name"]}
                for row in self.object_list.values("id", "name").order_by("name")
            ]
            results = sorted(results, key=lambda d: d["text"])
        else:
            results = None
        return {"results": results}

    def render_to_response(self, context, **response_kwargs):
        return self.render_to_json_response(context, **response_kwargs)



from http import HTTPStatus

from django.test import override_settings
from django.urls import reverse

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCorporationInfoFactory,
    UserFactory,
)
from app_utils.testing import NoSocketsTestCase

from structuretimers.admin import _get_multiselect_display
from structuretimers.models import NotificationRule, StagingSystem, Timer

from .testdata.factory import (
    DiscordWebhookFactory,
    EveSolarSystemFactory,
    NotificationRuleFactory,
    StagingSystemFactory,
)


class TestNotificationRule_ChangeList(NoSocketsTestCase):
    def test_can_open_page_normally(self):
        # given
        webhook = DiscordWebhookFactory()
        NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=webhook,
        )
        NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            require_timer_types=[Timer.Type.ARMOR],
            webhook=webhook,
        )
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=webhook,
        )
        rule.require_corporations.add(EveCorporationInfoFactory())
        NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            is_important=NotificationRule.Clause.EXCLUDED,
            webhook=webhook,
        )
        user = UserFactory(is_staff=True, is_superuser=True)
        self.client.force_login(user)

        # user tries to add new notification rule
        add_page = self.client.get(
            reverse("admin:structuretimers_notificationrule_changelist")
        )
        self.assertEqual(add_page.status_code, HTTPStatus.OK)


class TestNotificationRule_Validations(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.webhook = DiscordWebhookFactory()
        cls.add_url = reverse("admin:structuretimers_notificationrule_add")
        cls.changelist_url = reverse(
            "admin:structuretimers_notificationrule_changelist"
        )
        cls.form_data = {
            "is_important": NotificationRule.Clause.ANY,
            "is_opsec": NotificationRule.Clause.ANY,
            "ping_type": NotificationRule.PingType.NONE,
            "scheduled_time": NotificationRule.MINUTES_15,
            "trigger": NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            "webhook": cls.webhook.pk,
            "_save": "Save",
        }
        cls.user = UserFactory(is_staff=True, is_superuser=True)

    def setUp(self):
        self.client.force_login(self.user)

    def test_can_create_rule(self):
        # given
        form_data = self.form_data | {}

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertRedirects(response, self.changelist_url)
        self.assertEqual(NotificationRule.objects.count(), 1)

    def test_can_not_have_same_options_timer_types(self):
        # given
        form_data = self.form_data | {
            "require_timer_types": [Timer.Type.ANCHORING, Timer.Type.HULL],
            "exclude_timer_types": [Timer.Type.ANCHORING, Timer.Type.ARMOR],
        }

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(NotificationRule.objects.count(), 0)
        form = response.context["adminform"].form
        self.assertTrue(form.errors)

    def test_can_not_have_same_options_objectives(self):
        # given
        form_data = self.form_data | {
            "require_objectives": [Timer.Objective.FRIENDLY, Timer.Objective.HOSTILE],
            "exclude_objectives": [Timer.Objective.FRIENDLY, Timer.Objective.NEUTRAL],
        }

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(NotificationRule.objects.count(), 0)
        form = response.context["adminform"].form
        self.assertTrue(form.errors)

    def test_can_not_have_same_options_visibility(self):
        # given
        form_data = self.form_data | {
            "require_visibility": [Timer.Visibility.CORPORATION],
            "exclude_visibility": [Timer.Visibility.CORPORATION],
        }

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(NotificationRule.objects.count(), 0)
        form = response.context["adminform"].form
        self.assertTrue(form.errors)

    def test_can_not_have_same_options_corporations(self):
        # given
        corp_1 = EveCorporationInfoFactory()
        corp_2 = EveCorporationInfoFactory()

        form_data = self.form_data | {
            "require_corporations": [corp_1.pk, corp_2.pk],
            "exclude_corporations": [corp_1.pk],
        }

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(NotificationRule.objects.count(), 0)
        form = response.context["adminform"].form
        self.assertTrue(form.errors)

    def test_can_not_have_same_options_alliances(self):
        # given
        all_1 = EveAllianceInfoFactory()
        all_2 = EveAllianceInfoFactory()
        form_data = self.form_data | {
            "require_alliances": [all_1.pk, all_2.pk],
            "exclude_alliances": [all_1.pk],
        }

        # when
        response = self.client.post(self.add_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(NotificationRule.objects.count(), 0)
        form = response.context["adminform"].form
        self.assertTrue(form.errors)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestStagingSystemAdmin(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserFactory(is_staff=True, is_superuser=True)
        cls.url_add = reverse("admin:structuretimers_stagingsystem_add")

    def test_should_create_new_staging_system(self):
        # given
        self.client.force_login(self.user)
        solar_system = EveSolarSystemFactory()

        # when
        res = self.client.post(self.url_add, data={"eve_solar_system": solar_system.pk})

        # then
        self.assertEqual(res.status_code, HTTPStatus.FOUND)
        self.assertEqual(StagingSystem.objects.count(), 1)
        obj = StagingSystem.objects.first()
        self.assertEqual(obj.eve_solar_system, solar_system)
        self.assertFalse(obj.is_main)

    def test_should_ensure_only_one_obj_is_main(self):
        # given
        self.client.force_login(self.user)
        solar_system_1 = EveSolarSystemFactory()
        StagingSystemFactory(eve_solar_system=solar_system_1, is_main=True)
        solar_system_2 = EveSolarSystemFactory()

        # when
        res = self.client.post(
            self.url_add,
            data={"eve_solar_system": solar_system_2.pk, "is_main": True},
        )

        # then
        self.assertEqual(res.status_code, HTTPStatus.FOUND)
        self.assertEqual(
            StagingSystem.objects.filter(is_main=True).get().eve_solar_system,
            solar_system_2,
        )


class TestGetMultiselectDisplay(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.choices = [
            (1, "alpha"),
            (2, "bravo"),
        ]

    def test_returns_value_if_found(self):
        self.assertEqual(_get_multiselect_display(1, self.choices), "alpha")
        self.assertEqual(_get_multiselect_display(2, self.choices), "bravo")

    def test_raises_exception_if_not_found(self):
        with self.assertRaises(ValueError):
            _get_multiselect_display(3, self.choices)



"""Forms."""

import datetime as dt
from typing import Any, Dict

import puremagic
import requests

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from .constants import EveGroupId
from .models import Timer

logger = get_extension_logger(__name__)

DATETIME_FORMAT = "%Y-%m-%d %H:%M"


class TimerForm(forms.ModelForm):
    """Form for timers."""

    ASTERISK_HTML = mark_safe('<i class="fas fa-asterisk"></i>')
    TIME_REMAINING_WIDGET_ATTRS = {
        "class": "timer-time-remaining-field",
    }
    TIME_REMAINING_HELP_TEXT = _(
        "This field is calculated from the current time. "
        "Alternatively, you can enter the date above in the `Date` field."
    )
    eve_solar_system_2 = forms.CharField(
        required=True,
        label=format_html("{} {}", _("Solar System"), ASTERISK_HTML),
        widget=forms.Select(attrs={"class": "select2-solar-systems"}),
    )
    structure_type_2 = forms.CharField(
        required=True,
        label=format_html("{} {}", _("Structure Type"), ASTERISK_HTML),
        widget=forms.Select(attrs={"class": "select2-structure-types"}),
    )
    objective = forms.ChoiceField(
        initial=Timer.Objective.UNDEFINED,
        choices=Timer.Objective.choices,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    timer_type = forms.ChoiceField(
        choices=Timer.Type.choices,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    visibility = forms.ChoiceField(
        choices=Timer.Visibility.choices,
        widget=forms.Select(attrs={"class": "select2-render"}),
    )
    date = forms.DateTimeField(
        required=False,
        label=_("Date"),
        widget=forms.DateTimeInput(
            attrs={"id": "timer-date-field"}, format=DATETIME_FORMAT
        ),
        # input_formats=[DATETIME_FORMAT],
        help_text=_(
            "The date when the timer happens. "
            "Alternatively, you can enter the remaining time below."
        ),
    )
    days_left = forms.IntegerField(
        required=False,
        label=_("Days Remaining"),
        validators=[MinValueValidator(0)],
        widget=forms.NumberInput(attrs=TIME_REMAINING_WIDGET_ATTRS),
        help_text=TIME_REMAINING_HELP_TEXT,
    )
    hours_left = forms.IntegerField(
        required=False,
        label=_("Hours Remaining"),
        validators=[MinValueValidator(0), MaxValueValidator(23)],
        widget=forms.NumberInput(attrs=TIME_REMAINING_WIDGET_ATTRS),
        help_text=TIME_REMAINING_HELP_TEXT,
    )
    minutes_left = forms.IntegerField(
        required=False,
        label=_("Minutes Remaining"),
        validators=[MinValueValidator(0), MaxValueValidator(59)],
        widget=forms.NumberInput(attrs=TIME_REMAINING_WIDGET_ATTRS),
        help_text=TIME_REMAINING_HELP_TEXT,
    )
    details_image_url = forms.URLField(
        required=False,
        help_text=_("Paste a public URL to an image into this field."),
    )

    class Meta:
        model = Timer
        fields = (
            "eve_solar_system_2",
            "location_details",
            "structure_type_2",
            "timer_type",
            "structure_name",
            "owner_name",
            "objective",
            "date",
            "days_left",
            "hours_left",
            "minutes_left",
            "details_image_url",
            "details_notes",
            "visibility",
            "is_opsec",
            "is_important",
        )

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        if "instance" in kwargs and kwargs["instance"] is not None:
            my_instance = kwargs["instance"]
            self.is_new = False
        else:
            my_instance = None
            self.is_new = True

        super().__init__(*args, **kwargs)

        if my_instance:
            self.fields["eve_solar_system_2"].widget.choices = [
                (
                    str(my_instance.eve_solar_system_id),
                    my_instance.eve_solar_system.name,
                )
            ]
            self.fields["structure_type_2"].widget.choices = [
                (str(my_instance.structure_type_id), my_instance.structure_type.name)
            ]

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("eve_solar_system_2"):
            self._clean_solar_system(cleaned_data)

        if cleaned_data.get("structure_type_2"):
            self._clean_structure_type(cleaned_data)

        if cleaned_data.get("details_image_url"):
            self._clean_image(cleaned_data)

        days_left = cleaned_data.get("days_left")
        hours_left = cleaned_data.get("hours_left")
        minutes_left = cleaned_data.get("minutes_left")
        date = cleaned_data.get("date")
        if any([days_left, hours_left, minutes_left]):
            if days_left is None:
                days_left = cleaned_data["days_left"] = 0
            if hours_left is None:
                hours_left = cleaned_data["hours_left"] = 0
            if minutes_left is None:
                minutes_left = cleaned_data["minutes_left"] = 0

        timer_type = cleaned_data.get("timer_type")
        if (
            timer_type != Timer.Type.PRELIMINARY
            and days_left is None
            and hours_left is None
            and minutes_left is None
            and date is None
        ):
            cleaned_data["timer_type"] = Timer.Type.PRELIMINARY.value
        if timer_type == Timer.Type.PRELIMINARY and (
            days_left is not None
            or hours_left is not None
            or minutes_left is not None
            or date is not None
        ):
            cleaned_data["timer_type"] = Timer.Type.NONE.value

    def _clean_structure_type(self, cleaned_data: Dict[str, Any]):
        try:
            structure_type = EveType.objects.get(id=cleaned_data["structure_type_2"])
        except EveType.DoesNotExist as ex:
            raise ValidationError(
                {"structure_type_2": _("Invalid structure type.")}
            ) from ex

        self.fields["structure_type_2"].widget.choices = [
            (str(structure_type.id), structure_type.name)
        ]
        if (
            cleaned_data.get("timer_type") == Timer.Type.MOONMINING
            and structure_type.eve_group_id != EveGroupId.REFINERY
        ):
            raise ValidationError(
                {"timer_type": _("Moon mining timers are valid for refineries only.")}
            )
        if (
            cleaned_data.get("timer_type") == Timer.Type.THEFT
            and structure_type.eve_group_id != EveGroupId.SKYHOOK
        ):
            raise ValidationError(
                {"timer_type": _("Theft timers are valid for skyhook only.")}
            )

    def _clean_solar_system(self, cleaned_data: Dict[str, Any]):
        try:
            solar_system = EveSolarSystem.objects.get(
                id=cleaned_data["eve_solar_system_2"]
            )
        except EveSolarSystem.DoesNotExist as ex:
            raise ValidationError(
                {"eve_solar_system_2": _("Invalid solar system.")}
            ) from ex
        self.fields["eve_solar_system_2"].widget.choices = [
            (str(solar_system.id), solar_system.name)
        ]

    @staticmethod
    def _clean_image(cleaned_data: Dict[str, Any]):
        details_image_url = cleaned_data.get("details_image_url", "")
        try:
            r = requests.get(details_image_url, stream=True, timeout=(3.0, 10.0))
            r.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as ex:
            logger.warning(
                "Failed to load image from URL: %s", details_image_url, exc_info=True
            )
            raise forms.ValidationError(
                {
                    "details_image_url": _(
                        "Failed to load image file. Please double check URL."
                    )
                },
                code="details_url_failed_to_load",
            ) from ex

        image_type = puremagic.from_string(r.content, mime=True)
        if image_type not in {"image/gif", "image/jpeg", "image/png"}:
            logger.warning(
                "%s is not a valid image type for URL: %s",
                image_type,
                details_image_url,
            )
            raise forms.ValidationError(
                {
                    "details_image_url": _(
                        _(
                            "URL does not point to a valid image file. "
                            "Valid types are: gif, jpeg, png"
                        )
                    )
                },
                code="details_url_unsupported_type",
            )

    def save(self, commit=True):
        timer = super().save(commit=False)

        # character / corporation / alliance
        if self.is_new:
            character = self.user.profile.main_character
            try:
                alliance = character.alliance
            except EveAllianceInfo.DoesNotExist:
                alliance = EveAllianceInfo.objects.create_alliance(
                    character.alliance_id
                )
            try:
                corporation = character.corporation
            except EveCorporationInfo.DoesNotExist:
                corporation = EveCorporationInfo.objects.create_corporation(
                    character.corporation_id
                )
            logger.debug(
                (
                    "Determined timer save request is on behalf of "
                    "character %s corporation %s"
                ),
                character,
                corporation,
            )
            timer.eve_character = character
            timer.eve_corporation = corporation
            timer.eve_alliance = alliance
            timer.user = self.user

        # calculate future time
        days_left = self.cleaned_data.get("days_left")
        hours_left = self.cleaned_data.get("hours_left")
        minutes_left = self.cleaned_data.get("minutes_left")
        date = self.cleaned_data.get("date")
        if date is not None:
            timer.date = date
        elif (
            days_left is not None
            and hours_left is not None
            and minutes_left is not None
        ):
            future_time = dt.timedelta(
                days=days_left, hours=hours_left, minutes=minutes_left
            )
            current_time = now()
            date = current_time + future_time
            logger.debug(
                "Determined timer eve time is %s - current time %s, adding %s",
                date,
                current_time,
                future_time,
            )
            timer.date = date
        else:
            timer.date = None

        # structure type
        timer.structure_type_id = self.cleaned_data.get("structure_type_2")
        timer.eve_solar_system_id = self.cleaned_data.get("eve_solar_system_2")

        if commit:
            timer.save()
        return timer



"""Settings."""

from app_utils.app_settings import clean_setting

STRUCTURETIMERS_MAX_AGE_FOR_NOTIFICATIONS = clean_setting(
    "STRUCTURETIMERS_MAX_AGE_FOR_NOTIFICATIONS", 60
)
"""Will not schedule notifications for timers,
which have elapsed more than x minutes ago.
"""

STRUCTURETIMERS_NOTIFICATIONS_ENABLED = clean_setting(
    "STRUCTURETIMERS_NOTIFICATIONS_ENABLED", True
)
"""Whether notifications for timers are scheduled at all."""

STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS = clean_setting(
    "STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS", default_value=30, min_value=1
)
"""Minimum age in days for a timer to be considered obsolete.
Obsolete timers will automatically be deleted.
"""

STRUCTURETIMERS_DEFAULT_PAGE_LENGTH = clean_setting(
    "STRUCTURETIMERS_DEFAULT_PAGE_LENGTH", 10
)
"""Default page size for timerboard.
Must be an integer value from the current options as seen in the app.
"""

STRUCTURETIMERS_PAGING_ENABLED = clean_setting("STRUCTURETIMERS_PAGING_ENABLED", True)
"""Whether paging is enabled on the timerboard."""

STRUCTURETIMER_NOTIFICATION_SET_AVATAR = clean_setting(
    "STRUCTURETIMER_NOTIFICATION_SET_AVATAR", True
)
"""Whether structures sets the name and avatar icon of a webhook."""



from django.core.management.base import BaseCommand, CommandError
from django.db.models import QuerySet
from django.utils.timezone import now
from eveuniverse.models import EveSolarSystem

from allianceauth.timerboard.models import Timer as AuthTimer
from app_utils.datetime import DATETIME_FORMAT
from app_utils.django import app_labels

from structuretimers.constants import EveTypeId
from structuretimers.models import Timer

_OBJECTIVE_MAP = {
    AuthTimer.Objective.FRIENDLY: Timer.Objective.FRIENDLY,
    AuthTimer.Objective.HOSTILE: Timer.Objective.HOSTILE,
    AuthTimer.Objective.NEUTRAL: Timer.Objective.NEUTRAL,
}


_STRUCTURE_MAP = {
    AuthTimer.Structure.ANSIBLEX: EveTypeId.ANSIBLEX,
    AuthTimer.Structure.ASTRAHUS: EveTypeId.ASTRAHUS,
    AuthTimer.Structure.ATHANOR: EveTypeId.ATHANOR,
    AuthTimer.Structure.AZBEL: EveTypeId.AZBEL,
    AuthTimer.Structure.FORTIZAR: EveTypeId.FORTIZAR,
    AuthTimer.Structure.IHUB: EveTypeId.IHUB,
    AuthTimer.Structure.KEEPSTAR: EveTypeId.KEEPSTAR,
    AuthTimer.Structure.MERCDEN: EveTypeId.MERCENARY_DEN,
    AuthTimer.Structure.METENOX: EveTypeId.METENOX_MOON_DRILL,
    AuthTimer.Structure.MOONPOP: EveTypeId.ATHANOR,
    AuthTimer.Structure.ORBITALSKYHOOK: EveTypeId.ORBITAL_SKYHOOK,
    AuthTimer.Structure.PHAROLUX: EveTypeId.PHAROLUX,
    AuthTimer.Structure.POCO: EveTypeId.CUSTOMS_OFFICE,
    AuthTimer.Structure.POSL: EveTypeId.CALDARI_CONTROL_TOWER,
    AuthTimer.Structure.POSM: EveTypeId.CALDARI_CONTROL_TOWER_MEDIUM,
    AuthTimer.Structure.POSS: EveTypeId.CALDARI_CONTROL_TOWER_SMALL,
    AuthTimer.Structure.RAITARU: EveTypeId.RAITARU,
    AuthTimer.Structure.SOTIYO: EveTypeId.SOTIYO,
    AuthTimer.Structure.TATARA: EveTypeId.TATARA,
    AuthTimer.Structure.TENEBREX: EveTypeId.TENEBREX,
}

_TIMER_TYPE_MAP = {
    AuthTimer.TimerType.ANCHORING: Timer.Type.ANCHORING,
    AuthTimer.TimerType.ARMOR: Timer.Type.ARMOR,
    AuthTimer.TimerType.FINAL: Timer.Type.FINAL,
    AuthTimer.TimerType.HULL: Timer.Type.HULL,
    AuthTimer.TimerType.THEFT: Timer.Type.THEFT,
    AuthTimer.TimerType.UNANCHORING: Timer.Type.UNANCHORING,
    AuthTimer.TimerType.UNSPECIFIED: Timer.Type.NONE,
}


class Command(BaseCommand):
    help = "Migrate pending timers from Alliance Auth's Structure Timers app."

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Perform's a dry run. Does not actually migrate, but will show potential issues.",
        )

    def _migrate_timers(
        self, auth_timers_qs: QuerySet[AuthTimer], is_dry_run: bool
    ) -> None:
        migrated_count = 0
        issues_count = 0
        skipped_count = 0
        for auth_timer in auth_timers_qs:
            try:
                eve_solar_system = EveSolarSystem.objects.get(name=auth_timer.system)
            except EveSolarSystem.DoesNotExist:
                self.stdout.write(
                    self.style.WARNING(
                        f"Can't migrate timer with unrecognized solar system: {_auth_timer_str(auth_timer)}"
                    )
                )
                issues_count += 1
                continue

            if auth_timer.structure not in _STRUCTURE_MAP:
                self.stdout.write(
                    self.style.WARNING(
                        f"Can't migrate timer with unrecognized structure: {_auth_timer_str(auth_timer)}"
                    )
                )
                issues_count += 1
                continue

            structure_type_id = _STRUCTURE_MAP[auth_timer.structure]

            if auth_timer.objective not in _OBJECTIVE_MAP:
                self.stdout.write(
                    self.style.NOTICE(
                        f"Unrecognized objective for timer: {_auth_timer_str(auth_timer)}"
                    )
                )
                objective = Timer.Objective.UNDEFINED
            else:
                objective = _OBJECTIVE_MAP[auth_timer.objective]

            if auth_timer.timer_type not in _TIMER_TYPE_MAP:
                self.stdout.write(
                    self.style.NOTICE(
                        f"Unrecognized type for timer: {_auth_timer_str(auth_timer)}"
                    )
                )
                timer_type = Timer.Type.NONE
            else:
                timer_type = _TIMER_TYPE_MAP[auth_timer.timer_type]

            if (
                auth_timer.structure == AuthTimer.Structure.MOONPOP
                and auth_timer.timer_type == AuthTimer.TimerType.UNSPECIFIED
            ):
                timer_type = Timer.Type.MOONMINING

            if auth_timer.corp_timer:
                visibility = Timer.Visibility.CORPORATION
            else:
                visibility = Timer.Visibility.UNRESTRICTED

            try:
                Timer.objects.get(
                    date=auth_timer.eve_time,
                    eve_solar_system=eve_solar_system,
                    structure_type_id=structure_type_id,
                    timer_type=timer_type,
                )

            except Timer.DoesNotExist:
                if not is_dry_run:
                    Timer.objects.create(
                        date=auth_timer.eve_time,
                        details_notes=auth_timer.details,
                        eve_alliance=auth_timer.eve_corp.alliance,
                        eve_character=auth_timer.eve_character,
                        eve_corporation=auth_timer.eve_corp,
                        eve_solar_system=eve_solar_system,
                        is_important=auth_timer.important,
                        location_details=auth_timer.planet_moon,
                        objective=objective,
                        structure_type_id=structure_type_id,
                        timer_type=timer_type,
                        user=auth_timer.user,
                        visibility=visibility,
                    )
                migrated_count += 1

            else:
                self.stdout.write(
                    self.style.NOTICE(
                        f"Skipping already existing timer: {_auth_timer_str(auth_timer)}"
                    )
                )
                skipped_count += 1

        total = migrated_count + issues_count + skipped_count
        self.stdout.write(
            f"Results: Migrated: {migrated_count} - Issues: {issues_count} "
            f"- Total: {total}"
        )

    def handle(self, *args, **options):
        if "timerboard" not in app_labels():
            raise CommandError("The Alliance Auth timerboard app is not installed.")

        auth_timers_qs = AuthTimer.objects.filter(eve_time__gt=now())
        is_dry_run = options["dry_run"]
        count = auth_timers_qs.count()
        topic = "dry run" if is_dry_run else "migration"
        self.stdout.write(f"Started {topic} for {count} timers")
        self._migrate_timers(auth_timers_qs, is_dry_run=is_dry_run)
        self.stdout.write(self.style.SUCCESS(f"{topic.capitalize()} completed!"))


def _auth_timer_str(timer: AuthTimer) -> str:
    eve_time = timer.eve_time.strftime(DATETIME_FORMAT)
    timer_type = AuthTimer.TimerType(timer.timer_type).label
    location = timer.system
    if timer.planet_moon:
        location += f" - {timer.planet_moon}"
    return f"{timer_type} in {location} @ {eve_time}"



from unittest.mock import Mock, patch

from requests.exceptions import ConnectionError as NewConnectionError
from requests.exceptions import HTTPError

from app_utils.testing import NoSocketsTestCase

from structuretimers.constants import EveTypeId
from structuretimers.forms import TimerForm
from structuretimers.models import Timer
from structuretimers.tests.testdata.factory import (
    CitadelTypeFactory,
    EveSolarSystemLowSecFactory,
    RefineryTypeFactory,
    SkyhookTypeFactory,
    UserWithAccessFactory,
)

from .testdata import test_image_filename

FORMS_PATH = "structuretimers.forms"
MODELS_PATH = "structuretimers.models"


def bytes_from_file(filename, chunksize=8192):
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(chunksize)
            if chunk:
                for b in chunk:
                    yield b
            else:
                break


def create_form_data(**kwargs):
    form_data = {
        "eve_solar_system_2": 30004984,
        "structure_type_2": EveTypeId.ASTRAHUS.value,
        "timer_type": Timer.Type.NONE,
        "objective": Timer.Objective.UNDEFINED,
        "visibility": Timer.Visibility.UNRESTRICTED,
    }
    if kwargs:
        form_data.update(kwargs)
    return form_data


class TestTimerFormIsValid(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.system_abune = EveSolarSystemLowSecFactory(id=30004984, name="Abune")
        cls.type_astrahus = CitadelTypeFactory(id=35832, name="Astrahus")
        cls.type_athanor = RefineryTypeFactory(id=35835, name="Athanor")
        SkyhookTypeFactory()

    def test_should_accept_normal_timer_with_date_parts(self):
        # given
        form_data = create_form_data(days_left=0, hours_left=3, minutes_left=30)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())

    def test_should_accept_normal_timer_with_date(self):
        # given
        form_data = create_form_data(date="2022-03-05 20:07")
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())

    def test_should_accept_normal_timer_with_partial_date_1(self):
        # given
        form_data = create_form_data(days_left=1)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)

    def test_should_accept_normal_timer_with_partial_date_2(self):
        # given
        form_data = create_form_data(hours_left=1)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)

    def test_should_accept_normal_timer_with_partial_date_3(self):
        # given
        form_data = create_form_data(minutes_left=1)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)

    def test_should_accept_preliminary_timer_without_date(self):
        # given
        form_data = create_form_data(timer_type=Timer.Type.PRELIMINARY)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())

    def test_should_upgrade_preliminary_timer_when_date_parts_specified(self):
        # given
        form_data = create_form_data(
            timer_type=Timer.Type.PRELIMINARY,
            days_left=0,
            hours_left=3,
            minutes_left=30,
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)
        self.assertIsNone(form.cleaned_data["date"])
        self.assertIsNotNone(form.cleaned_data["days_left"])
        self.assertIsNotNone(form.cleaned_data["hours_left"])
        self.assertIsNotNone(form.cleaned_data["minutes_left"])

    def test_should_upgrade_preliminary_timer_when_date_parts_specified_2(self):
        # given
        form_data = create_form_data(timer_type=Timer.Type.PRELIMINARY, days_left=5)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)
        self.assertIsNone(form.cleaned_data["date"])
        self.assertIsNotNone(form.cleaned_data["days_left"])
        self.assertIsNotNone(form.cleaned_data["hours_left"])
        self.assertIsNotNone(form.cleaned_data["minutes_left"])

    def test_should_upgrade_preliminary_timer_when_date_specified(self):
        # given
        form_data = create_form_data(
            timer_type=Timer.Type.PRELIMINARY, date="2022-03-05 20:07"
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.NONE)
        self.assertIsNone(form.cleaned_data["days_left"])
        self.assertIsNone(form.cleaned_data["hours_left"])
        self.assertIsNone(form.cleaned_data["minutes_left"])
        self.assertIsNotNone(form.cleaned_data["date"])

    def test_should_set_timer_as_preliminary_timer_when_no_date_specified(self):
        # given
        form_data = create_form_data(timer_type=Timer.Type.ARMOR)
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["timer_type"], Timer.Type.PRELIMINARY)
        self.assertIsNone(form.cleaned_data["date"])
        self.assertIsNone(form.cleaned_data["days_left"])
        self.assertIsNone(form.cleaned_data["hours_left"])
        self.assertIsNone(form.cleaned_data["minutes_left"])

    def test_should_not_accept_timer_without_solar_system(self):
        # given
        form_data = create_form_data(days_left=0, hours_left=3, minutes_left=30)
        del form_data["eve_solar_system_2"]
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    def test_should_not_accept_timer_without_structure_type(self):
        # given
        form_data = create_form_data(days_left=0, hours_left=3, minutes_left=30)
        del form_data["structure_type_2"]
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    def test_should_not_accept_invalid_days(self):
        # given
        form_data = create_form_data(days_left=-1, hours_left=3, minutes_left=30)
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    def test_should_not_accept_invalid_date(self):
        # given
        form_data = create_form_data(date="2022.31.05 20:07:59")
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    def test_should_not_accept_moon_mining_type_for_non_mining_structures(self):
        # given
        form_data = create_form_data(
            timer_type=Timer.Type.MOONMINING,
            structure_type_2=self.type_astrahus.id,
            days_left=0,
            hours_left=3,
            minutes_left=30,
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    @patch(FORMS_PATH + ".requests.get", spec=True)
    def test_should_create_timer_with_valid_details_image(self, mock_get):
        # given
        image_file = bytearray(bytes_from_file(test_image_filename()))
        mock_get.return_value.content = image_file
        form_data = create_form_data(
            days_left=0,
            hours_left=3,
            minutes_left=30,
            details_image_url="http://www.example.com/image.png",
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertTrue(form.is_valid())

    @patch(FORMS_PATH + ".requests.get", spec=True)
    def test_should_not_allow_invalid_link_for_detail_images(self, mock_get):
        # given
        image_file = bytearray(bytes_from_file(test_image_filename()))
        mock_get.return_value.content = image_file
        form_data = create_form_data(
            days_left=0,
            hours_left=3,
            minutes_left=30,
            details_image_url="invalid-url",
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    @patch(FORMS_PATH + ".requests.get", spec=True)
    def test_should_show_error_when_image_can_not_be_loaded_1(self, mock_get):
        # given
        mock_get.side_effect = NewConnectionError
        form_data = create_form_data(
            days_left=0,
            hours_left=3,
            minutes_left=30,
            details_image_url="http://www.example.com/image.png",
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    @patch(FORMS_PATH + ".requests.get", spec=True)
    def test_should_show_error_when_image_can_not_be_loaded_2(self, mock_get):
        # given
        mock_get.side_effect = HTTPError
        form_data = create_form_data(
            days_left=0,
            hours_left=3,
            minutes_left=30,
            details_image_url="http://www.example.com/image.png",
        )
        form = TimerForm(data=form_data)
        # when / then
        self.assertFalse(form.is_valid())

    def test_should_allow_theft_timer_for_skyhook_only(self):
        cases = [
            ("skyhook", EveTypeId.ORBITAL_SKYHOOK.value, True),
            ("citadel", EveTypeId.ASTRAHUS.value, False),
            ("poco", EveTypeId.CUSTOMS_OFFICE.value, False),
            ("ihub", EveTypeId.IHUB.value, False),
            ("tcu", EveTypeId.TCU.value, False),
        ]
        for tc in cases:
            with self.subTest(name=tc[0]):
                form_data = create_form_data(
                    days_left=0,
                    hours_left=3,
                    minutes_left=30,
                    timer_type=Timer.Type.THEFT,
                    structure_type_2=tc[1],
                )
                form = TimerForm(data=form_data)
                self.assertIs(form.is_valid(), tc[2], form.errors)


@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
@patch(MODELS_PATH + "._task_schedule_notifications_for_timer", Mock())
class TestTimerFormSave(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserWithAccessFactory()
        cls.system_abune = EveSolarSystemLowSecFactory(id=30004984, name="Abune")
        cls.type_astrahus = CitadelTypeFactory(id=35832, name="Astrahus")
        cls.type_athanor = RefineryTypeFactory(id=35835, name="Athanor")
        SkyhookTypeFactory()

    def test_should_create_new_normal_timer(self):
        # given
        form_data = create_form_data(
            days_left=0, hours_left=3, minutes_left=30, timer_type=Timer.Type.ARMOR
        )
        form = TimerForm(user=self.user, data=form_data)
        # when
        form.save()
        # then
        timer = Timer.objects.first()
        self.assertEqual(timer.timer_type, Timer.Type.ARMOR)
        self.assertIsNotNone(timer.date)

    def test_should_create_new_preliminary_timer(self):
        # given
        form_data = create_form_data()
        form = TimerForm(user=self.user, data=form_data)
        # when
        form.save()
        # then
        timer = Timer.objects.first()
        self.assertEqual(timer.timer_type, Timer.Type.PRELIMINARY)
        self.assertIsNone(timer.date)

    def test_should_promote_preliminary_timer_to_normal_timer(self):
        # given
        form_data = create_form_data(timer_type=Timer.Type.PRELIMINARY, days_left=1)
        form = TimerForm(user=self.user, data=form_data)
        # when
        form.save()
        # then
        timer = Timer.objects.first()
        self.assertEqual(timer.timer_type, Timer.Type.NONE)
        self.assertIsNotNone(timer.date)



import datetime as dt
from http import HTTPStatus
from typing import Optional, Set

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.urls import reverse
from django.utils.timezone import now

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
)
from app_utils.testing import (
    NoSocketsTestCase,
    json_response_to_dict,
    json_response_to_python,
)

from structuretimers.models import Timer
from structuretimers.tests.testdata.factory import (
    CitadelTypeFactory,
    DistancesFromStagingFactory,
    EveSolarSystemFactory,
    StagingSystemFactory,
    TimerFactory,
    UserMainFactory,
    UserWithAccessFactory,
    UserWithManageFactory,
)

MODELS_PATH = "structuretimers.models"


class TestViewBase(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_1 = UserWithAccessFactory()
        cls.user_2 = UserWithManageFactory()
        cls.user_3 = UserWithAccessFactory()

        cls.timer_1 = TimerFactory(
            structure_name="Timer 1",
            location_details="Near the star",
            date=now() + dt.timedelta(hours=4),
            eve_character=cls.character_1,
            eve_corporation=cls.corporation_1,
            user=cls.user_1,
            eve_solar_system=cls.system_abune,
            structure_type=cls.type_astrahus,
            owner_name="Big Boss",
            details_image_url="http://www.example.com/dummy.png",
            details_notes="Some notes",
        )
        cls.timer_2 = TimerFactory(
            structure_name="Timer 2",
            date=now() - dt.timedelta(hours=8),
            eve_character=cls.character_1,
            eve_corporation=cls.corporation_1,
            user=cls.user_1,
            eve_solar_system=cls.system_abune,
            structure_type=cls.type_raitaru,
            is_important=True,
        )
        cls.timer_3 = TimerFactory(
            structure_name="Timer 3",
            date=now() - dt.timedelta(hours=8),
            eve_character=cls.character_1,
            eve_corporation=cls.corporation_1,
            user=cls.user_1,
            eve_solar_system=cls.system_enaluri,
            structure_type=cls.type_astrahus,
        )
        cls.timer_4 = TimerFactory(
            timer_type=Timer.Type.PRELIMINARY,
            structure_name="Timer 4",
            eve_character=cls.character_1,
            eve_corporation=cls.corporation_1,
            user=cls.user_1,
        )


class TestTimerList_SelectedStagingSystem(NoSocketsTestCase):
    def test_should_open_with_main_staging_system(self):
        # given
        StagingSystemFactory()
        staging_system = StagingSystemFactory(is_main=True)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response.context_data["selected_staging_system"], staging_system
        )

    def test_should_open_with_first_staging_system_when_there_is_no_main(self):
        # given
        staging_system = StagingSystemFactory()
        StagingSystemFactory()
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response.context_data["selected_staging_system"], staging_system
        )

    def test_should_open_without_staging_system(self):
        # given
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIsNone(response.context_data["selected_staging_system"])

    def test_should_ignore_wrong_staging_system_name(self):
        # given
        staging_system = StagingSystemFactory(is_main=True)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/?staging=invalid_name")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response.context_data["selected_staging_system"], staging_system
        )

    def test_should_handle_multiple_invalid_staging_systems(self):
        # given
        StagingSystemFactory(eve_solar_system=None)
        StagingSystemFactory(eve_solar_system=None)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIsNone(response.context_data["selected_staging_system"])

    def test_should_handle_multiple_invalid_staging_systems_with_main(self):
        # given
        StagingSystemFactory(eve_solar_system=None, is_main=True)
        StagingSystemFactory(eve_solar_system=None)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get("/structuretimers/")

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIsNone(response.context_data["selected_staging_system"])


def json_response_pks(response: HttpResponse) -> Set[int]:
    return set(json_response_to_dict(response).keys())


class TestTimerListData(NoSocketsTestCase):
    view_name = "structuretimers:timer_list_data"

    def _get_timer_list_data(
        self, tab_name: str = "current", user: Optional[User] = None
    ):
        if not user:
            user = self.user_1
        self.client.force_login(user)
        return self.client.get(
            reverse("structuretimers:timer_list_data", args=[tab_name])
        )

    def _get_timer_list_data_ids(
        self, tab_name: str = "current", user: Optional[User] = None
    ) -> set:
        response = self._get_timer_list_data(tab_name, user)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        return set(json_response_to_dict(response).keys())

    # def test_timer_list_view_loads(self):
    #     request = self.factory.get(reverse("structuretimers:timer_list"))
    #     request.user = self.user_1
    #     response = views.timer_list(request)
    #     self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_return_current_timers(self):
        # given
        timer = TimerFactory()
        TimerFactory(date=now() - dt.timedelta(hours=8))
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_return_past_timers(self):
        # given
        TimerFactory()
        timer = TimerFactory(date=now() - dt.timedelta(hours=8))
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=["past"]))

        # then
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_return_preliminary_timers(self):
        # given
        TimerFactory()
        timer = TimerFactory(timer_type=Timer.Type.PRELIMINARY)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=["preliminary"]))

        # then
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_not_give_access_without_basic_permission(self):
        # given
        TimerFactory()
        self.client.force_login(UserMainFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_should_return_corp_restricted_for_corp_member_only(self):
        # given
        user = UserWithAccessFactory()
        ec = EveCorporationInfoFactory(
            corporation_id=user.profile.main_character.corporation_id
        )
        timer = TimerFactory(
            visibility=Timer.Visibility.CORPORATION, eve_corporation=ec
        )
        TimerFactory(visibility=Timer.Visibility.CORPORATION)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_show_alliance_restricted_to_alliance_member(self):
        # given
        corp = EveCorporationInfoFactory(create_alliance=True)
        ec = EveCharacterFactory(
            corporation_id=corp.corporation_id, alliance_id=corp.alliance.alliance_id
        )
        user = UserWithAccessFactory(main_character__character=ec)
        timer = TimerFactory(
            visibility=Timer.Visibility.ALLIANCE, eve_alliance=corp.alliance
        )
        TimerFactory(visibility=Timer.Visibility.ALLIANCE)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_show_opsec_restricted__when_user_has_opsec_permission(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structuretimers.basic_access",
                "structuretimers.opsec_access",
            ]
        )
        timer = TimerFactory(is_opsec=True)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_not_show_opsec_restricted__when_user_not_has_opsec_permission(self):
        # given
        user = UserMainFactory(
            permissions__=[
                "structuretimers.basic_access",
            ]
        )
        TimerFactory(is_opsec=True)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), set())

    def test_should_return_corp_timer_to_creator_when_in_different_corp(self):
        # given
        user = UserWithAccessFactory()
        timer = TimerFactory(
            visibility=Timer.Visibility.CORPORATION,
            eve_corporation=EveCorporationInfoFactory(),
            user=user,
        )
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_return_corp_timer_to_creator_when_in_different_alliance(self):
        # given
        user = UserWithAccessFactory()
        timer = TimerFactory(
            visibility=Timer.Visibility.ALLIANCE,
            eve_alliance=EveAllianceInfoFactory(),
            user=user,
        )
        self.client.force_login(user)

        # when
        response = self.client.get(reverse(self.view_name, args=["current"]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertSetEqual(json_response_pks(response), {timer.id})

    def test_should_include_distances_1(self):
        # given
        timer = TimerFactory()
        distances = DistancesFromStagingFactory(timer=timer, light_years=1.2, jumps=3)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(
            reverse(self.view_name, args=["current"])
            + f"?staging={distances.staging_system.pk}"
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_dict(response)
        obj = data[timer.id]
        self.assertEqual(obj["distance_light_years"], 1.2)
        self.assertEqual(obj["distance_jumps"], 3)
        self.assertTrue(obj["distance"])


class TestDetailView(NoSocketsTestCase):
    view_name = "structuretimers:detail"

    def test_should_show_normal_timer(self):
        # given
        timer = TimerFactory(structure_name="Alpha")
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "structuretimers/timer_detail.html")
        self.assertIn("Alpha", response.rendered_content)

    def test_should_show_preliminary_timer(self):
        # given
        timer = TimerFactory(structure_name="Alpha", timer_type=Timer.Type.PRELIMINARY)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name, args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "structuretimers/timer_detail.html")
        self.assertIn("Alpha", response.rendered_content)

    def test_should_not_show_details_when_no_access(self):
        # given
        timer = TimerFactory(structure_name="Alpha", is_opsec=True)
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse("structuretimers:detail", args=[timer.pk]))
        # then
        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)


class TestSelect2Views_SolarSystems(NoSocketsTestCase):
    view_name = "structuretimers:select2_solar_systems"

    def test_should_return_solar_systems(self):
        # given
        EveSolarSystemFactory(id=30004984, name="Abune")
        EveSolarSystemFactory()
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name), data={"term": "abu"})

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        self.assertEqual(data, {"results": [{"id": 30004984, "text": "Abune"}]})

    def test_should_return_empty_solar_system_list(self):
        # given
        EveSolarSystemFactory(id=30004984, name="Abune")
        EveSolarSystemFactory()
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse(self.view_name))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        self.assertEqual(data, {"results": None})


class TestSelect2Views_StructureTypes(NoSocketsTestCase):
    view_name = "structuretimers:select2_structure_types"

    def test_should_return_structure_types(self):
        # given
        CitadelTypeFactory(id=35832, name="Astrahus")
        CitadelTypeFactory()
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(
            reverse("structuretimers:select2_structure_types"), data={"term": "ast"}
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        self.assertEqual(data, {"results": [{"id": 35832, "text": "Astrahus"}]})

    def test_should_return_empty_structure_types_list(self):
        # given
        CitadelTypeFactory(id=35832, name="Astrahus")
        CitadelTypeFactory()
        self.client.force_login(UserWithAccessFactory())

        # when
        response = self.client.get(reverse("structuretimers:select2_structure_types"))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        self.assertEqual(data, {"results": None})



"""Global constants."""

# pylint: disable=missing-class-docstring

from enum import IntEnum


class EveCategoryId(IntEnum):
    ORBITAL = 46
    STRUCTURE = 65


class EveGroupId(IntEnum):
    CITADEL = 1657
    CONTROL_TOWER = 365
    MERCENARY_DEN = 4810
    MOBILE_DEPOT = 1246
    PIRATE_FORWARD_OPERATING_BASE = 4644
    REFINERY = 1406
    SKYHOOK = 4736


class EveTypeId(IntEnum):
    ANSIBLEX = 35841
    ASTRAHUS = 35832
    ATHANOR = 35835
    AZBEL = 35826
    CALDARI_CONTROL_TOWER = 16213
    CALDARI_CONTROL_TOWER_MEDIUM = 20061
    CALDARI_CONTROL_TOWER_SMALL = 20062
    CUSTOMS_OFFICE = 2233
    FORTIZAR = 35833
    IHUB = 32458
    KEEPSTAR = 35834
    MERCENARY_DEN = 85230
    METENOX_MOON_DRILL = 81826
    ORBITAL_SKYHOOK = 81080
    PHAROLUX = 35840
    RAITARU = 35825
    SOTIYO = 35827
    STATION = 1529
    TATARA = 35836
    TCU = 32226
    TENEBREX = 37534



# Generated by Django 5.2.14 on 2026-05-20 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structuretimers", "0006_add_theft_timer_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="timer",
            name="notification_rules",
            field=models.ManyToManyField(
                help_text="Notification rules conforming with this timer",
                through="structuretimers.ScheduledNotification",
                through_fields=("timer", "notification_rule"),
                to="structuretimers.notificationrule",
            ),
        ),
    ]



from django.core.management import call_command
from django.core.management.base import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from structuretimers import __title__
from structuretimers.constants import EveCategoryId, EveGroupId, EveTypeId

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = "Preloads data required for this app from ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            __title__,
            "--category_id",
            str(EveCategoryId.STRUCTURE.value),
            "--group_id",
            str(EveGroupId.CONTROL_TOWER.value),
            "--group_id",
            str(EveGroupId.MOBILE_DEPOT.value),
            "--group_id",
            str(EveGroupId.MERCENARY_DEN.value),
            "--group_id",
            str(EveGroupId.PIRATE_FORWARD_OPERATING_BASE.value),
            "--type_id",
            str(EveTypeId.CUSTOMS_OFFICE.value),
            "--type_id",
            str(EveTypeId.ORBITAL_SKYHOOK.value),
            "--type_id",
            str(EveTypeId.TCU.value),
            "--type_id",
            str(EveTypeId.IHUB.value),
        )



from contextlib import AbstractContextManager

from django.contrib.auth.models import User
from django.db import transaction

from allianceauth.tests.auth_utils import AuthUtils


class isolated_subtest(AbstractContextManager):
    """Guarantees database isolation for a subTest block by forcing
    an atomic savepoint to roll back on exit, regardless of whether
    the test passed, failed, or errored.

    To be used in tests created with `django.test.TestCase`.
    """

    def __enter__(self):
        self.atomic = transaction.atomic()
        self.atomic.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            transaction.set_rollback(True)
        finally:
            return self.atomic.__exit__(exc_type, exc_val, exc_tb)


def add_permission_to_user_by_name(perm: str, user: User) -> User:
    """adds permission to given user by name and returns updated user object"""
    AuthUtils.add_permission_to_user_by_name(perm, user)
    return User.objects.get(pk=user.pk)



"""An app for keeping track of Eve Online structure timers
with Alliance Auth and Discord.
"""

# pylint: disable = invalid-name
default_app_config = "structuretimers.apps.StructureTimersConfig"

__version__ = "3.1.0"
__title__ = "Structure Timers"



import datetime as dt
from unittest.mock import Mock, patch

from celery import Task

from django.test import TestCase, TransactionTestCase
from django.utils.timezone import now

from structuretimers.models import NotificationRule, ScheduledNotification, Timer
from structuretimers.tasks import (
    calc_timer_distances_for_all_staging_systems,
    housekeeping,
    notify_about_new_timer,
    schedule_notifications_for_rule,
    schedule_notifications_for_timer,
    send_messages_for_webhook,
    send_scheduled_notification,
)
from structuretimers.tests.testdata.factory import (
    DiscordWebhookFactory,
    NotificationRuleFactory,
    ScheduledNotificationFactory,
    StagingSystemFactory,
    TimerFactory,
)

MODULE_PATH = "structuretimers.tasks"


# class TestCaseBase(TestCase):
#     @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
#     def setUp(self) -> None:
#         webhook = DiscordWebhookFactory()
#         webhook.clear_queue()
#         rule = NotificationRuleFactory(
#             trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
#             scheduled_time=NotificationRule.MINUTES_15,
#             webhook=webhook,
#         )
#         timer = TimerFactory(
#             structure_name="Test_1",
#             eve_solar_system=self.system_abune,
#             structure_type=self.type_raitaru,
#             date=now() + dt.timedelta(minutes=30),
#         )


@patch(MODULE_PATH + ".DiscordWebhook.send_queued_messages", spec=True)
@patch(MODULE_PATH + ".logger", spec=True)
class TestSendMessagesForWebhook(TestCase):
    def test_normal(self, mock_logger, mock_send_queued_messages):
        # given
        webhook = DiscordWebhookFactory()

        # when
        send_messages_for_webhook(webhook.pk)

        # then
        self.assertEqual(mock_send_queued_messages.call_count, 1)
        self.assertEqual(mock_logger.info.call_count, 2)
        self.assertEqual(mock_logger.error.call_count, 0)

    def test_disabled_webhook(self, mock_logger, mock_send_queued_messages):
        # given
        webhook = DiscordWebhookFactory(is_enabled=False)

        # when
        send_messages_for_webhook(webhook.pk)

        # then
        self.assertEqual(mock_send_queued_messages.call_count, 0)
        self.assertEqual(mock_logger.info.call_count, 1)
        self.assertEqual(mock_logger.error.call_count, 0)


@patch(MODULE_PATH + ".notify_about_new_timer", spec=True)
@patch(MODULE_PATH + ".send_scheduled_notification", spec=True)
class TestScheduleNotificationForTimer(TestCase):
    def test_should_schedule_new_notification_for_time_reached(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        timer = TimerFactory()
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED
        )

        # when
        schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

        # then
        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(timer.scheduled_notifications.filter(notification_rule=rule))

    def test_should_not_create_notification_for_preliminary_timer(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        timer = TimerFactory(timer_type=Timer.Type.PRELIMINARY)
        NotificationRuleFactory(trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED)

        # when/then
        with self.assertRaises(ValueError):
            schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

    def test_should_remove_old_notifications(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        timer: Timer = TimerFactory()
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED
        )
        notification_old = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            timer_date=timer.date + dt.timedelta(minutes=5),
            notification_date=timer.date - dt.timedelta(minutes=5),
            celery_task_id="99",
        )

        # when
        schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

        # then
        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            timer.scheduled_notifications.filter(notification_rule=rule).exists()
        )
        self.assertFalse(
            ScheduledNotification.objects.filter(pk=notification_old.pk).exists()
        )

    def test_should_schedule_notification_for_new_timer(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.NEW_TIMER_CREATED
        )
        timer: Timer = TimerFactory()

        # when
        schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

        # then
        self.assertTrue(mock_send_notification_for_timer.apply_async.called)
        _, kwargs = mock_send_notification_for_timer.apply_async.call_args
        self.assertEqual(kwargs["kwargs"]["timer_pk"], timer.pk)
        self.assertEqual(kwargs["kwargs"]["notification_rule_pk"], rule.pk)

    def test_no_notification_for_new_timer_if_no_rule(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        timer: Timer = TimerFactory()

        # when
        schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

        # then
        self.assertFalse(mock_send_notification_for_timer.apply_async.called)

    def test_should_abort_when_outdated(
        self, mock_send_notification, mock_send_notification_for_timer
    ):
        # given
        timer: Timer = TimerFactory(date=now() - dt.timedelta(hours=1))

        # when
        schedule_notifications_for_timer(timer_pk=timer.pk, is_new=True)

        # then
        self.assertFalse(mock_send_notification_for_timer.apply_async.called)


@patch(MODULE_PATH + ".send_scheduled_notification", spec=True)
class TestScheduleNotificationForRule(TestCase):
    def test_should_schedule_new_notification_when_rule_matches(
        self, mock_send_notification
    ):
        # given
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        timer = TimerFactory()
        rule = NotificationRuleFactory()

        # when
        schedule_notifications_for_rule(rule.pk)

        # then
        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            timer.scheduled_notifications.filter(notification_rule=rule).exists()
        )

    def test_should_remove_old_notifications(self, mock_send_notification):
        # given
        timer: Timer = TimerFactory()
        rule = NotificationRuleFactory()
        mock_send_notification.apply_async.return_value.task_id = "my_task_id"
        notification_old = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            timer_date=timer.date + dt.timedelta(minutes=5),
            notification_date=timer.date - dt.timedelta(minutes=5),
            celery_task_id="99",
        )

        # when
        schedule_notifications_for_rule(rule.pk)

        # then
        self.assertTrue(mock_send_notification.apply_async.called)
        self.assertTrue(
            timer.scheduled_notifications.filter(notification_rule=rule).exists()
        )
        self.assertFalse(
            ScheduledNotification.objects.filter(pk=notification_old.pk).exists()
        )

    def test_abort_when_has_the_wrong_trigger(self, mock_send_notification):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.NEW_TIMER_CREATED
        )

        # when
        schedule_notifications_for_rule(rule.pk)

        # then
        self.assertFalse(mock_send_notification.apply_async.called)


@patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
@patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
class TestSendScheduledNotification(TransactionTestCase):
    def test_should_send_notification(self, mock_send_messages_for_webhook):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
        )
        timer = TimerFactory(
            structure_name="Test_1",
            date=now() + dt.timedelta(minutes=30),
        )
        scheduled_notification = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            celery_task_id="my-id-123",
            timer_date=now() + dt.timedelta(hours=1),
            notification_date=now() + dt.timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )

        # then
        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)

    def test_should_revoked_notification_when_incorrect_task_instance(
        self, mock_send_messages_for_webhook
    ):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
        )
        timer = TimerFactory(
            structure_name="Test_1",
            date=now() + dt.timedelta(minutes=30),
        )
        scheduled_notification = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            celery_task_id="my-id-123",
            timer_date=now() + dt.timedelta(hours=1),
            notification_date=now() + dt.timedelta(minutes=30),
        )
        mock_task = Mock(**{"request.id": "my-id-456"})

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)

    def test_discard_notification_when_rule_disabled(
        self, mock_send_messages_for_webhook
    ):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
            is_enabled=False,
        )
        timer = TimerFactory(
            structure_name="Test_1",
            date=now() + dt.timedelta(minutes=30),
        )
        scheduled_notification = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            celery_task_id="my-id-123",
            timer_date=now() + dt.timedelta(hours=1),
            notification_date=now() + dt.timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)

    def test_should_ignore_when_notification_was_deleted(
        self, mock_send_messages_for_webhook
    ):
        # given
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(mock_task, scheduled_notification_pk=666)

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)

    def test_should_abort_when_webhook_disabled(self, mock_send_messages_for_webhook):
        # given
        webhook = DiscordWebhookFactory(is_enabled=False)
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
            is_enabled=False,
            webhook=webhook,
        )
        timer = TimerFactory(
            structure_name="Test_1",
            date=now() + dt.timedelta(minutes=30),
        )
        scheduled_notification = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            celery_task_id="my-id-123",
            timer_date=now() + dt.timedelta(hours=1),
            notification_date=now() + dt.timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)

    def test_should_discard_when_timer_is_outdated(
        self, mock_send_messages_for_webhook
    ):
        # given
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_15,
            is_enabled=False,
        )
        timer = TimerFactory(
            structure_name="Test_1", date=now() - dt.timedelta(hours=1)
        )
        scheduled_notification = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            celery_task_id="my-id-123",
            timer_date=now() + dt.timedelta(hours=1),
            notification_date=now() + dt.timedelta(minutes=30),
        )
        mock_task = Mock(spec=Task)
        mock_task.request.id = "my-id-123"

        # when
        send_scheduled_notification_inner = (
            send_scheduled_notification.__wrapped__.__func__
        )
        send_scheduled_notification_inner(
            mock_task, scheduled_notification_pk=scheduled_notification.pk
        )

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)


@patch(MODULE_PATH + ".send_messages_for_webhook", spec=True)
class TestNotifyAboutNewTimer(TestCase):
    def test_should_notify_about_new_timer(self, mock_send_messages_for_webhook):
        # given
        webhook = DiscordWebhookFactory()
        timer = TimerFactory()
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.NEW_TIMER_CREATED, webhook=webhook
        )

        # when
        notify_about_new_timer(timer.pk, rule.pk)

        # then
        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)
        _, kwargs = mock_send_messages_for_webhook.apply_async.call_args
        self.assertListEqual(kwargs["args"], [webhook.pk])

    def test_should_notify_about_timer_when_rule_does_not_exist(
        self, mock_send_messages_for_webhook
    ):
        # given
        webhook = DiscordWebhookFactory()
        timer: Timer = TimerFactory()
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED, webhook=webhook
        )

        # when
        notify_about_new_timer(timer.pk, rule.pk)

        # then
        self.assertTrue(mock_send_messages_for_webhook.apply_async.called)
        _, kwargs = mock_send_messages_for_webhook.apply_async.call_args
        self.assertListEqual(kwargs["args"], [webhook.pk])

    def test_should_abort_when_rule_disabled(self, mock_send_messages_for_webhook):
        # given
        timer = TimerFactory()
        rule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.NEW_TIMER_CREATED,
            is_enabled=False,
        )
        # when
        notify_about_new_timer(timer.pk, rule.pk)

        # then
        self.assertFalse(mock_send_messages_for_webhook.apply_async.called)


@patch(MODULE_PATH + ".Timer.objects.delete_obsolete", spec=True)
class TestHousekeeping(TestCase):
    def test_should_run_housekeeping(self, mock_delete_obsolete):
        # given
        mock_delete_obsolete.return_value = 1
        # when
        housekeeping()
        # then
        self.assertTrue(mock_delete_obsolete.called)


@patch(MODULE_PATH + ".calc_timer_distances_for_staging_system", spec=True)
class TestTimerDistancesForAllStagingSystems(TestCase):
    def test_should_calc_distances(self, mock_calc_timer_distances_for_staging_system):
        # given
        timer = TimerFactory(date=now() + dt.timedelta(minutes=30))
        StagingSystemFactory()

        # when
        calc_timer_distances_for_all_staging_systems(timer.pk)

        # then
        self.assertEqual(
            mock_calc_timer_distances_for_staging_system.apply_async.call_count, 1
        )



import datetime as dt
from unittest.mock import patch

from django.db.models import QuerySet
from django.utils.timezone import now

from app_utils.testdata_factories import EveCorporationInfoFactory
from app_utils.testing import NoSocketsTestCase, queryset_pks

from structuretimers.models import NotificationRule, Timer
from structuretimers.tests.testdata.factory import (
    DiscordWebhookFactory,
    NotificationRuleFactory,
    TimerFactory,
)

MODELS_PATH = "structuretimers.models"


class TestTimer_QuerySet(NoSocketsTestCase):
    def test_should_match_one_timer(self):
        # given
        timer_1 = TimerFactory(timer_type=Timer.Type.ARMOR)
        TimerFactory(timer_type=Timer.Type.HULL)
        rule = NotificationRuleFactory(require_timer_types=[Timer.Type.ARMOR])

        # when
        qs: QuerySet = Timer.objects.all().conforms_with_notification_rule(rule)

        # then
        self.assertSetEqual(queryset_pks(qs), {timer_1.pk})

    def test_should_match_no_timer(self):
        # given
        TimerFactory(eve_corporation=EveCorporationInfoFactory())
        rule = NotificationRuleFactory()
        rule.require_corporations.add(EveCorporationInfoFactory())

        # when
        qs = Timer.objects.all().conforms_with_notification_rule(rule)

        # then
        self.assertFalse(qs.exists())

    def test_should_match_both_timers(self):
        # given
        timer_1 = TimerFactory(objective=Timer.Objective.FRIENDLY)
        timer_2 = TimerFactory(objective=Timer.Objective.FRIENDLY)
        rule = NotificationRuleFactory(require_objectives=[Timer.Objective.FRIENDLY])

        # when
        qs = Timer.objects.all().conforms_with_notification_rule(rule)

        # then
        self.assertSetEqual(queryset_pks(qs), {timer_1.pk, timer_2.pk})


class TestTimer_Manger(NoSocketsTestCase):
    @patch("structuretimers.managers.STRUCTURETIMERS_TIMERS_OBSOLETE_AFTER_DAYS", 1)
    def test_delete_old_timer(self):
        timer_1 = TimerFactory(date=now())
        TimerFactory(date=now() - dt.timedelta(days=1, seconds=1))
        result = Timer.objects.delete_obsolete()
        self.assertEqual(result, 1)
        self.assertSetEqual(queryset_pks(Timer.objects.all()), {timer_1.pk})

    def test_can_handle_no_timers(self):
        result = Timer.objects.delete_obsolete()
        self.assertEqual(result, 0)


class TestNotificationRule_QuerySet(NoSocketsTestCase):
    def test_should_contain_matching_rules_only(self):
        # given
        webhook = DiscordWebhookFactory()
        rule_1 = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=10,
            require_timer_types=[Timer.Type.ARMOR],
            webhook=webhook,
        )
        rule_2 = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=15,
            require_timer_types=[Timer.Type.ARMOR],
            webhook=webhook,
        )
        NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=15,
            require_objectives=[Timer.Objective.FRIENDLY],
            webhook=webhook,
        )
        timer = TimerFactory(
            structure_name="Test Timer",
            date=now() + dt.timedelta(hours=4),
            timer_type=Timer.Type.ARMOR,
            objective=Timer.Objective.HOSTILE,
        )

        # when
        qs = NotificationRule.objects.all().conforms_with_timer(timer)

        # then
        self.assertSetEqual(queryset_pks(qs), {rule_1.pk, rule_2.pk})

    def test_should_not_match_any_rule(self):
        # given
        webhook = DiscordWebhookFactory()
        NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=15,
            require_objectives=[Timer.Objective.FRIENDLY],
            webhook=webhook,
        )
        timer = TimerFactory(
            structure_name="Test Timer",
            date=now() + dt.timedelta(hours=4),
            timer_type=Timer.Type.HULL,
            objective=Timer.Objective.HOSTILE,
        )

        # when
        qs = NotificationRule.objects.all().conforms_with_timer(timer)

        # then
        self.assertSetEqual(queryset_pks(qs), set())
